from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import cast

import stim

from tqecd.exceptions import TQECDException, TQECDWarning
from tqecd.pauli import PauliString
from tqecd.predicates import (
    does_not_contain_both_reset_and_measurement,
    is_valid_input_circuit,
)
from tqecd.utils import (
    collapse_pauli_strings_at_moment,
    has_circuit_repeat_block,
    has_measurement,
    has_only_reset_or_is_virtual,
    has_reset,
    is_virtual_moment,
    iter_stim_circuit_by_moments,
)


class Fragment:
    def __init__(self, circuit: stim.Circuit):
        """A sub-circuit guaranteed to end with a moment filled by measurement
        instructions.

        Fragment instances represent sub-circuits that contain:

        1. zero or more moments composed of `reset` and any other instructions
            except measurement instructions,
        2. zero or more moments composed of "computation" instructions (anything
            that is not a measurement or a reset),
        3. one moment composed of `measurement` and any other instructions
            except reset instructions.

        Raises:
            TQECDException: if the provided `stim.Circuit` instance contains a
                stim.CircuitRepeatBlock instance.
            TQECDException: if any moment from the provided circuit contains
                both a reset and a measurement operation.
            TQECDException: if the provided circuit does not end with at least
                one measurement.

        Args:
            circuit: the circuit represented by the instance.
        """
        if has_circuit_repeat_block(circuit):
            raise TQECDException(
                "Breaking invariant: Cannot initialise a Fragment with a "
                "stim.CircuitRepeatBlock instance but found one. Did you "
                "meant to use FragmentLoop?"
            )
        # The line below has no type issue as the circuit does not contain
        # any stim.CircuitRepeatBlock instance, and so iter_stim_circuit_by_moments
        # can only return stim.Circuit instances.
        moments = [
            cast(stim.Circuit, moment).copy()
            for moment in iter_stim_circuit_by_moments(circuit)
        ]

        self._circuit = circuit
        self._resets: list[PauliString] = []
        self._measurements: list[PauliString] = []

        for moment in moments:
            if is_virtual_moment(moment):
                continue
            if not has_reset(moment):
                break
            if not does_not_contain_both_reset_and_measurement(moment):
                raise TQECDException(
                    "Breaking invariant: found a moment with both reset "
                    f"and measurement operations:\n{moment}"
                )
            self._resets.extend(collapse_pauli_strings_at_moment(moment))

        for moment in reversed(moments):
            if is_virtual_moment(moment):
                continue
            if not has_measurement(moment):
                break
            # Insert new measurement at the front to keep them correctly ordered.
            self._measurements = (
                collapse_pauli_strings_at_moment(moment) + self._measurements
            )

        if not self._measurements:
            raise TQECDException(
                "A Fragment should end with at least one measurement. "
                "The provided circuit does not seem to check that condition.\n"
                f"Provided circuit:\n{circuit}"
            )

    @property
    def resets(self) -> list[PauliString]:
        """Get the reset instructions at the front on the Fragment.

        Returns:
            all the reset instructions that appear at the beginning of the represented
            circuit, in the order of appearance, and in increasing qubit order for resets
            that are performed in parallel.
        """
        return self._resets

    @property
    def measurements(self) -> list[PauliString]:
        """Get the measurement instructions at the back on the Fragment.

        Returns:
            all the measurement instructions that appear at the end of the represented
            circuit, in the order of appearance, and in increasing qubit order for
            measurements that are performed in parallel.
        """
        return self._measurements

    @property
    def measurements_qubits(self) -> list[int]:
        return [measurement.qubit for measurement in self.measurements]

    @property
    def num_measurements(self) -> int:
        return len(self._measurements)

    @property
    def circuit(self) -> stim.Circuit:
        return self._circuit

    def get_tableau(self) -> stim.Tableau:
        return self._circuit.to_tableau(
            ignore_measurement=True, ignore_noise=True, ignore_reset=True
        )

    def __repr__(self) -> str:
        return f"Fragment(circuit={self._circuit!r})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Fragment) and self._circuit == other._circuit


@dataclass(frozen=True)
class FragmentLoop:
    fragments: list[Fragment | FragmentLoop]
    repetitions: int

    def __post_init__(self) -> None:
        if self.repetitions < 1:
            raise TQECDException(
                "Cannot have a FragmentLoop with 0 or less repetitions."
            )
        if not self.fragments:
            raise TQECDException(
                "Cannot initialise a FragmentLoop instance without any "
                "fragment for the loop body."
            )

    def with_repetitions(self, repetitions: int) -> FragmentLoop:
        return FragmentLoop(fragments=self.fragments, repetitions=repetitions)

    def __repr__(self) -> str:
        return f"FragmentLoop(repetitions={self.repetitions}, fragments={self.fragments!r})"


def _get_fragment_loop(repeat_block: stim.CircuitRepeatBlock) -> FragmentLoop:
    try:
        body_fragments = split_stim_circuit_into_fragments(repeat_block.body_copy())
    except TQECDException as e:
        raise TQECDException(
            f"Error when splitting the following REPEAT block:\n{repeat_block.body_copy()}"
        ) from e
    return FragmentLoop(fragments=body_fragments, repetitions=repeat_block.repeat_count)


def split_stim_circuit_into_fragments(
    circuit: stim.Circuit,
) -> list[Fragment | FragmentLoop]:
    """Split the circuit into fragments.

    The provided circuit should check a few pre-conditions:

    - If there is one measurement (resp. reset) instruction between two TICK
      annotation, then no reset (resp. measurement) instruction should appear
      between these two TICK.
    - The circuit should be (recursively if it contains one or more instance of
      `stim.CircuitRepeatBlock`) composed of a succession of layers that should
      have the same shape:

      - starts with zero or more moments containing reset and non-measurement operations,
      - continuing with zero or more moments containing any non-collapsing operation
        (i.e., anything except reset and measurement operations).
      - ends with one moment containing measurement and non-reset operations.

      For the above reasons, be careful with reset/measurement combined operations
      (e.g., the `stim` instruction `MR` that performs in one instruction a
      measurement and a reset in the Z basis). These instructions are not supported
      by the `tqec` library and it is up to the user to check that the input circuit
      does not contain combined measurements/resets instructions.

    Args:
        circuit (stim.Circuit): the circuit to split into Fragment instances.

    Raises:
        TQECDException: If the circuit contains at least one moment (i.e., group of
            instructions between two TICK annotations) that are composed of both
            measurement and reset instructions.
        TQECDException: If the circuit contains combined measurement/reset instructions.
        TQECDException: If the provided circuit could not be split into fragments due
            to an invalid structure.

    Returns:
        the resulting fragments.
    """
    potential_error_reason = is_valid_input_circuit(circuit)
    if potential_error_reason is not None:
        raise TQECDException(potential_error_reason)

    fragments: list[Fragment | FragmentLoop] = []
    current_fragment = stim.Circuit()

    moments_iterator = iter_stim_circuit_by_moments(circuit)
    for moment in moments_iterator:
        # If we have a REPEAT block
        if isinstance(moment, stim.CircuitRepeatBlock):
            # Purge the current fragment.
            # Note that the following lines should only be triggered on invalid
            # inputs as once one measurement is found, all the following measurements
            # are collected and a Fragment instance is created (see content of next
            # elif branch). So if we are here and there is some partially collected
            # fragment (i.e., current_fragment is not empty), it means that it is
            # not terminated by measurements, which will raise an error when Fragment
            # is constructed.
            if current_fragment:
                raise TQECDException(
                    "Trying to start a REPEAT block without a cleanly finished Fragment. "
                    "The following instructions were found preceding the REPEAT block:\n"
                    + "\n\t".join(f"{m}" for m in current_fragment)
                    + "\nbut these instructions do not form a valid Fragment."
                )
            # Recurse to produce the Fragment instances for the loop body.
            fragments.append(_get_fragment_loop(moment))

        # If this is a measurement moment
        # we add the full moment to the current fragment and start a new one.
        elif has_measurement(moment):
            current_fragment += moment
            fragments.append(Fragment(current_fragment.copy()))
            current_fragment.clear()

        # This is either a regular instruction or a reset moment. In any case,
        # just add it to the current fragment.
        else:
            current_fragment += moment

    # If current_fragment is not empty here, this means that the circuit did not finish
    # with a measurement. This is strange, so for the moment raise an exception.
    if current_fragment:
        if has_only_reset_or_is_virtual(current_fragment):
            warnings.warn(
                "Found left-over reset gates when splitting a circuit. Make "
                "sure that each reset (even resets from measurement/reset "
                "combined instruction) is eventually followed by a measurement. "
                f"Unprocessed fragment:\n{current_fragment}",
                TQECDWarning,
            )
        else:
            raise TQECDException(
                "Circuit splitting did not finish on a measurement. "
                f"Unprocessed fragment: \n{current_fragment}"
            )
    return fragments



from __future__ import annotations

import typing as ty
from dataclasses import dataclass

from tqecd.boundary import BoundaryStabilizer
from tqecd.exceptions import TQECDException
from tqecd.fragment import Fragment, FragmentLoop
from tqecd.match_utils.cover import find_commuting_cover_on_target_qubits_sat
from tqecd.measurement import get_relative_measurement_index
from tqecd.pauli import PauliString, pauli_product


def _anti_commuting_stabilizers_indices(flows: list[BoundaryStabilizer]) -> list[int]:
    return [i for i in range(len(flows)) if flows[i].has_anticommuting_operations]


def _try_merge_anticommuting_flows_inplace(flows: list[BoundaryStabilizer]) -> None:
    """Merge as much anti-commuting flows as possible from the provided flows.

    This function try to merge together several :class:`BoundaryStabilizer`
    instances that anti-commute with their collapsing operations and provided
    in `flows`. It **modifies in-place the provided parameter**, removing
    anti-commuting flows and replacing them with the resulting commuting
    flow when found.

    Args:
        flows: a list of flows that might or might not contains flows that
            anti-commute with its collapsing operations.

    Raises:
        TQECDException: if the provided flows have different collapsing
            operations, hinting that they are not part of the same boundary,
            in which case it makes no sense to try to merge them together.
    """
    # Filtering out commuting operations as they cannot make anti-commuting
    # operations commuting.
    anti_commuting_index_to_flows_index: list[int] = (
        _anti_commuting_stabilizers_indices(flows)
    )

    # Early exit if there are no anti-commuting collapsing operations
    if not anti_commuting_index_to_flows_index:
        return

    collapsing_operations: list[set[PauliString]] = [
        set(flows[fi].collapsing_operations)
        for fi in anti_commuting_index_to_flows_index
    ]
    # Checking that all the provided flows are defined on the same boundary.
    # This is checked by comparing the collapsing operations for each
    # anti-commuting stabilizer and asserting that they are all equal.
    for i in range(1, len(collapsing_operations)):
        if collapsing_operations[0] != collapsing_operations[i]:
            raise TQECDException(
                "Cannot merge anti-commuting flows defined on different collapsing "
                "operations. Found the following difference:\nFlow 0 has the "
                "collapsing operations:\n\t"
                + "\n\t".join(f"- {c}" for c in collapsing_operations[0])
                + f"\nFlow {i} has the collapsing operations:\n\t"
                + "\n\t".join(f"- {c}" for c in collapsing_operations[i])
                + "\n"
            )

    # Computation of the Pauli string representing all the collapsing operations.
    # The goal of this method will be to find a cover from the provided flows such
    # as the resulting propagated Pauli string commutes with this one.
    collapsing_pauli = pauli_product(collapsing_operations[0])

    # Now, we want to find flows in anticommuting_stabilizers that, when taken
    # into account together, commute with collapsing_pauli.
    anticommuting_stabilizers: list[PauliString] = [
        flows[fi].before_collapse for fi in anti_commuting_index_to_flows_index
    ]
    indices_of_anti_commuting_stabilizers_to_merge = (
        find_commuting_cover_on_target_qubits_sat(
            collapsing_pauli, anticommuting_stabilizers
        )
    )
    # While there are anti-commuting stabilizers that can be merged.
    while indices_of_anti_commuting_stabilizers_to_merge is not None:
        # Recover all the stabilizers that should be merged together.
        flows_indices_of_stabilizers_to_merge = [
            anti_commuting_index_to_flows_index[i]
            for i in indices_of_anti_commuting_stabilizers_to_merge
        ]
        stabilizers_to_merge: list[BoundaryStabilizer] = [
            flows[i] for i in flows_indices_of_stabilizers_to_merge
        ]
        # Update the flows by removing the entries related to stabilizers that
        # will be merged and re-compute the anti-commuting stabilizers and map.
        for i in sorted(flows_indices_of_stabilizers_to_merge, reverse=True):
            flows.pop(i)
        anti_commuting_index_to_flows_index = _anti_commuting_stabilizers_indices(flows)
        anticommuting_stabilizers = [
            flows[fi].before_collapse for fi in anti_commuting_index_to_flows_index
        ]
        # Compute the resulting commuting stabilizer.
        new_commuting_stabilizer = stabilizers_to_merge[0]
        for removed_stabilizer in stabilizers_to_merge[1:]:
            new_commuting_stabilizer = new_commuting_stabilizer.merge(
                removed_stabilizer
            )
        # 3. Add the resulting commuting stabilizer to the flows.
        flows.append(new_commuting_stabilizer)
        # Update for loop condition
        indices_of_anti_commuting_stabilizers_to_merge = (
            find_commuting_cover_on_target_qubits_sat(
                collapsing_pauli, anticommuting_stabilizers
            )
        )


@dataclass
class FragmentFlows:
    """Stores stabilizer flows for a :class:`Fragment` instance.

    Attributes:
        creation: stabilizer flows that are created by the :class:`Fragment`.
            These flows originate from a single reset instruction contained
            in the :class:`Fragment` instance.
        destruction: stabilizer flows that end in the :class:`Fragment`. These
            flows are generated by propagating backwards the Pauli string stabilized
            by a measurement operation contained in the :class:`Fragment`.
        total_number_of_measurements: the total number of measurements contained
            in the represented :class:`Fragment`. Might be used to offset measurement
            offsets by this amount when the measurement is located on a :class:`Fragment`
            instance before the one represented by self.
    """

    creation: list[BoundaryStabilizer]
    destruction: list[BoundaryStabilizer]
    total_number_of_measurements: int

    @property
    def all_flows(self) -> ty.Iterator[BoundaryStabilizer]:
        yield from self.creation
        yield from self.destruction

    def remove_creation(self, index: int) -> None:
        self.creation.pop(index)

    def remove_destruction(self, index: int) -> None:
        self.destruction.pop(index)

    def remove_creations(self, indices: ty.Iterable[int]) -> None:
        for i in sorted(indices, reverse=True):
            self.remove_creation(i)

    def remove_destructions(self, indices: ty.Iterable[int]) -> None:
        for i in sorted(indices, reverse=True):
            self.remove_destruction(i)

    def without_trivial_flows(self) -> FragmentFlows:
        return FragmentFlows(
            creation=[bs for bs in self.creation if bs.is_trivial()],
            destruction=[bs for bs in self.destruction if bs.is_trivial()],
            total_number_of_measurements=self.total_number_of_measurements,
        )

    def try_merge_anticommuting_flows(self) -> None:
        _try_merge_anticommuting_flows_inplace(self.creation)
        _try_merge_anticommuting_flows_inplace(self.destruction)


@dataclass
class FragmentLoopFlows:
    """Store stabilizer flows for a FragmentLoop instance.

    This class is currently quite dumb and does not provide a sufficient
    API for generic stabilizer matching, but is enough for detectors
    that only include measurements from the current round and from the
    previous round.
    """

    fragment_flows: list[FragmentFlows | FragmentLoopFlows]
    repeat: int

    @property
    def creation(self) -> list[BoundaryStabilizer]:
        return self.fragment_flows[-1].creation

    @property
    def destruction(self) -> list[BoundaryStabilizer]:
        return self.fragment_flows[0].destruction

    @property
    def all_flows(self) -> ty.Iterator[BoundaryStabilizer]:
        yield from self.creation
        yield from self.destruction

    @property
    def total_number_of_measurements(self) -> int:
        return sum(flow.total_number_of_measurements for flow in self.fragment_flows)

    def remove_creation(self, index: int) -> None:
        self.creation.pop(index)

    def remove_destruction(self, index: int) -> None:
        self.destruction.pop(index)

    def remove_creations(self, indices: ty.Iterable[int]) -> None:
        for i in sorted(indices, reverse=True):
            self.remove_creation(i)

    def remove_destructions(self, indices: ty.Iterable[int]) -> None:
        for i in sorted(indices, reverse=True):
            self.remove_destruction(i)

    def try_merge_anticommuting_flows(self) -> None:
        _try_merge_anticommuting_flows_inplace(self.creation)
        _try_merge_anticommuting_flows_inplace(self.destruction)


def build_flows_from_fragments(
    fragments: ty.Sequence[Fragment | FragmentLoop],
) -> list[FragmentFlows | FragmentLoopFlows]:
    """Compute and return the stabilizer flows of the provided fragments.

    This function ensures that the returned list will have the same "shape"
    as the input one. In more details, that means that the following property
    should be checked (recursively if there is any FragmentLoop instance in
    the provided fragments):

    .. code-block:: python

        fragments: list[Fragment | FragmentLoop] = []  # anything here
        flows = build_flows_from_fragments(fragments)
        for frag, flow in zip(fragments, flows):
            assert (isinstance(frag, Fragment) and isinstance(flow, FragmentFlow)) or (
                isinstance(frag, FragmentLoop) and isinstance(flow, FragmentLoopFlow)
            )

    Args:
        fragments: the fragments composing the circuit to study and for which this
            function should compute flows.

    Returns:
        the computed flows for each of the provided fragments.
    """
    return [
        _build_flows_from_fragment(fragment)
        if isinstance(fragment, Fragment)
        else _build_flows_from_fragment_loop(fragment)
        for fragment in fragments
    ]


def _build_flows_from_fragment(fragment: Fragment) -> FragmentFlows:
    tableau = fragment.get_tableau()
    targets = list(range(len(tableau)))
    sorted_qubit_involved_in_measurements = fragment.measurements_qubits

    # First compute the flows created within the Fragment (i.e., originating from
    # reset instructions).
    creation_flows: list[BoundaryStabilizer] = []
    for reset_stabilizer in fragment.resets:
        final_stabilizer = reset_stabilizer.after(tableau, targets)
        involved_measurements_offsets = [
            get_relative_measurement_index(
                sorted_qubit_involved_in_measurements, m.qubit
            )
            for m in fragment.measurements
            if final_stabilizer.overlaps(m)
        ]
        involved_resets_qubits = [reset_stabilizer.qubit]
        creation_flows.append(
            BoundaryStabilizer(
                final_stabilizer,
                fragment.measurements,
                involved_measurements_offsets,
                frozenset(involved_resets_qubits),
                forward=True,
            )
        )

    # Then, compute the flows destructed by the Fragment (i.e., if that flow is
    # given as input, a set of measurements from the Fragment will commute with
    # the entire flow and collapse it to "no flow").
    tableau_inv = tableau.inverse()
    destruction_flows: list[BoundaryStabilizer] = []
    for measurement in fragment.measurements:
        if measurement.non_trivial_pauli_count != 1:
            raise TQECDException(
                "Found a measurement applied on several qubits. "
                "This is not implemented (yet?)."
            )
        initial_stabilizer = measurement.after(tableau_inv, targets)
        involved_measurements_offsets = [
            get_relative_measurement_index(
                sorted_qubit_involved_in_measurements, measurement.qubit
            )
        ]
        involved_resets_qubits = [
            m.qubit for m in fragment.resets if initial_stabilizer.overlaps(m)
        ]
        destruction_flows.append(
            BoundaryStabilizer(
                initial_stabilizer,
                fragment.resets,
                involved_measurements_offsets,
                frozenset(involved_resets_qubits),
                forward=False,
            ),
        )

    return FragmentFlows(
        creation=creation_flows,
        destruction=destruction_flows,
        total_number_of_measurements=len(fragment.measurements),
    )


def _build_flows_from_fragment_loop(fragment_loop: FragmentLoop) -> FragmentLoopFlows:
    return FragmentLoopFlows(
        build_flows_from_fragments(fragment_loop.fragments), fragment_loop.repetitions
    )



from typing import Union

import numpy
import pytest

from tqecd.boundary import BoundaryStabilizer, manhattan_distance
from tqecd.exceptions import TQECDException
from tqecd.measurement import RelativeMeasurementLocation
from tqecd.pauli import PauliString


def test_boundary_stabilizer_construction() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    Z0 = PauliString({0: "Z"})
    Z1 = PauliString({1: "Z"})

    BoundaryStabilizer(X0Z1, [Z0, Z1], [], frozenset([0]), True)
    BoundaryStabilizer(X0Z1, [Z0, Z1], [], frozenset([0]), False)
    BoundaryStabilizer(
        X0Z1, [], [RelativeMeasurementLocation(-1, 0)], frozenset([0]), True
    )
    BoundaryStabilizer(X0Z1, [Z0, Z0], [], frozenset([1, 4, 6]), True)
    BoundaryStabilizer(X0Z1, [Z0, Z0], [], frozenset([]), True)


def test_has_anti_commuting_operations() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    X0 = PauliString({0: "X"})
    Z0 = PauliString({0: "Z"})
    Z1 = PauliString({1: "Z"})

    stab = BoundaryStabilizer(X0Z1, [X0, Z1], [], frozenset([0]), True)
    assert not stab.has_anticommuting_operations

    stab = BoundaryStabilizer(X0Z1, [Z0, Z1], [], frozenset([0]), True)
    assert stab.has_anticommuting_operations

    stab = BoundaryStabilizer(X0Z1, [], [], frozenset([0]), True)
    assert not stab.has_anticommuting_operations


def test_boundary_stabilizer_collapsing_operations() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    X0 = PauliString({0: "X"})
    Z0 = PauliString({0: "Z"})
    Z1 = PauliString({1: "Z"})

    stab = BoundaryStabilizer(X0Z1, [X0, Z1], [], frozenset([0]), True)
    assert set(stab.collapsing_operations) == {X0, Z1}

    stab = BoundaryStabilizer(X0Z1, [Z0, Z1], [], frozenset([0]), True)
    assert set(stab.collapsing_operations) == {Z0, Z1}

    stab = BoundaryStabilizer(X0Z1, [], [], frozenset([0]), True)
    assert not list(stab.collapsing_operations)


def test_after_and_before_collapse() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    X0 = PauliString({0: "X"})
    Z0 = PauliString({0: "Z"})
    Z1 = PauliString({1: "Z"})

    stab = BoundaryStabilizer(X0Z1, [X0, Z1], [], frozenset([0]), True)
    assert stab.before_collapse == X0Z1
    assert stab.after_collapse == PauliString({})

    stab = BoundaryStabilizer(X0Z1, [Z0, Z1], [], frozenset([0]), True)
    assert stab.before_collapse == X0Z1
    with pytest.raises(TQECDException):
        stab.after_collapse

    stab = BoundaryStabilizer(X0Z1, [], [], frozenset([0]), True)
    assert stab.before_collapse == X0Z1
    assert stab.after_collapse == X0Z1


def test_merge() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    Z0Z1 = PauliString({0: "Z", 1: "Z"})
    Z0X1 = PauliString({0: "Z", 1: "X"})
    Y0 = PauliString({0: "Y"})
    X0 = PauliString({0: "X"})
    Z0 = PauliString({0: "Z"})
    Z1 = PauliString({1: "Z"})

    a = BoundaryStabilizer(
        X0Z1,
        [X0, Z1],
        [RelativeMeasurementLocation(-2, 0), RelativeMeasurementLocation(-1, 1)],
        frozenset([0]),
        True,
    )
    b = BoundaryStabilizer(
        Z0Z1, [X0, Z1], [RelativeMeasurementLocation(-1, 1)], frozenset([1]), True
    )
    c = BoundaryStabilizer(Z0X1, [Z0, Z1], [], frozenset([0]), True)
    d = BoundaryStabilizer(
        Z0X1, [X0, Z1], [RelativeMeasurementLocation(-20, 11)], frozenset([2]), False
    )
    ab = a.merge(b)
    assert ab.before_collapse == Y0
    assert set(ab.collapsing_operations) == {X0, Z1}
    assert set(ab.measurements) == {RelativeMeasurementLocation(-2, 0)}
    assert set(ab.resets_qubits) == {0, 1}

    with pytest.raises(TQECDException):
        a.merge(c)
    with pytest.raises(TQECDException):
        a.merge(d)


def test_with_measurement_offset() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    X0 = PauliString({0: "X"})
    Z1 = PauliString({1: "Z"})

    a = BoundaryStabilizer(
        X0Z1,
        [X0, Z1],
        [RelativeMeasurementLocation(-28, 3), RelativeMeasurementLocation(-10, 3)],
        frozenset([0]),
        True,
    )

    assert set(a.with_measurement_offset(9).measurements) == {
        RelativeMeasurementLocation(-19, 3),
        RelativeMeasurementLocation(-1, 3),
    }
    assert set(a.with_measurement_offset(-1).measurements) == {
        RelativeMeasurementLocation(-29, 3),
        RelativeMeasurementLocation(-11, 3),
    }
    with pytest.raises(
        TQECDException,
        match=r"^Relative measurement offsets should be strictly negative\.$",
    ):
        a.with_measurement_offset(10)


def test_coordinates() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    X0 = PauliString({0: "X"})
    Z1 = PauliString({1: "Z"})

    a = BoundaryStabilizer(
        X0Z1,
        [X0, Z1],
        [RelativeMeasurementLocation(-28, 3), RelativeMeasurementLocation(-10, 3)],
        frozenset([3]),
        True,
    )
    qubit_coordinates = (1.0, 2.0)
    numpy.testing.assert_allclose(
        a.coordinates({3: qubit_coordinates}), qubit_coordinates
    )


def test_coordinates_raises_error_if_invalid_qubit_mapping() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    X0 = PauliString({0: "X"})
    Z1 = PauliString({1: "Z"})

    a = BoundaryStabilizer(
        X0Z1,
        [X0, Z1],
        [RelativeMeasurementLocation(-28, 5), RelativeMeasurementLocation(-10, 3)],
        frozenset([3, 5]),
        True,
    )
    qubit_coordinates: dict[int, tuple[Union[float, int], ...]] = {3: (0, 1)}
    with pytest.raises(
        TQECDException,
        match=r"^Qubit index 5 required for detector assignment, but it does not have a valid QUBIT_COORDS statement\.$",
    ):
        a.coordinates(qubit_coordinates=qubit_coordinates)


def test_manhattan_distance() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    Z0X1 = PauliString({0: "Z", 1: "X"})
    X0 = PauliString({0: "X"})
    Z1 = PauliString({1: "Z"})

    qubits_coordinates: dict[int, tuple[float, ...]] = {
        i: (2.0 * i, i / 2) for i in range(10)
    }
    a = BoundaryStabilizer(
        X0Z1,
        [X0, Z1],
        [RelativeMeasurementLocation(-28, 3), RelativeMeasurementLocation(-10, 3)],
        frozenset([3]),
        True,
    )
    b = BoundaryStabilizer(
        Z0X1, [X0, Z1], [RelativeMeasurementLocation(-20, 6)], frozenset([6]), True
    )
    numpy.testing.assert_allclose(a.coordinates(qubits_coordinates), (6.0, 1.5))
    numpy.testing.assert_allclose(b.coordinates(qubits_coordinates), (12.0, 3.0))
    numpy.testing.assert_allclose(manhattan_distance(a, a, qubits_coordinates), 0.0)
    numpy.testing.assert_allclose(manhattan_distance(b, b, qubits_coordinates), 0.0)
    numpy.testing.assert_allclose(manhattan_distance(a, b, qubits_coordinates), 7.5)
    numpy.testing.assert_allclose(manhattan_distance(b, a, qubits_coordinates), 7.5)



from __future__ import annotations

from typing import cast

import stim

from tqecd.exceptions import TQECDException
from tqecd.flow import build_flows_from_fragments
from tqecd.fragment import Fragment, FragmentLoop, split_stim_circuit_into_fragments
from tqecd.match import MatchedDetector, match_detectors_from_flows_shallow
from tqecd.predicates import is_valid_input_circuit
from tqecd.utils import remove_duplicate_detectors


def _detectors_to_circuit(
    detectors: list[MatchedDetector], additional_coordinates: list[float] | None = None
) -> stim.Circuit:
    """Transform a list of detectors into a circuit.

    Args:
        detectors: detectors that will be included in the returned circuit.

    Returns:
        A ``stim.Circuit`` instance containing all the provided detectors.
    """
    if additional_coordinates is None:
        additional_coordinates = []

    circuit = stim.Circuit()

    for detector in detectors:
        circuit.append(detector.to_instruction())

    return circuit


def _shift_time_instruction(number_of_spatial_coordinates: int) -> stim.Circuit:
    args = tuple(0.0 for _ in range(number_of_spatial_coordinates)) + (1.0,)
    circuit = stim.Circuit()
    circuit.append(
        stim.CircuitInstruction("SHIFT_COORDS", targets=[], gate_args=list(args))
    )
    return circuit


def annotate_detectors_automatically(circuit: stim.Circuit) -> stim.Circuit:
    """Insert detectors into the provided circuit instance.

    This is the main user-facing function to automatically insert detectors into
    a quantum circuit composed of Clifford operations. The provided circuit should
    check some pre-conditions to be accepted, which are detailed in details below.

    First and foremost, the provided circuit should check the pre-conditions listed
    in the documentation of :func:`split_stim_circuit_into_fragments`.

    Args:
        circuit: circuit to insert detectors in.

    Returns:
        A new ``stim.Circuit`` instance with automatically computed detectors.
    """
    potential_error_reason = is_valid_input_circuit(circuit)
    if potential_error_reason is not None:
        raise TQECDException(potential_error_reason)

    fragments = split_stim_circuit_into_fragments(circuit)
    qubit_coords_map: dict[int, tuple[float, ...]] = {
        q: tuple(coords) for q, coords in circuit.get_final_qubit_coordinates().items()
    }
    return compile_fragments_to_circuit_with_detectors(fragments, qubit_coords_map)


def compile_fragments_to_circuit(
    fragments: list[Fragment | FragmentLoop],
) -> stim.Circuit:
    circuit = stim.Circuit()

    for fragment in fragments:
        if isinstance(fragment, Fragment):
            circuit += fragment.circuit
        else:  # isinstance(fragment, FragmentLoop):
            loop_body = compile_fragments_to_circuit(fragment.fragments)
            circuit += loop_body * fragment.repetitions
    return circuit


def _tick_circuit() -> stim.Circuit:
    return stim.Circuit("TICK")


def _insert_before_last_tick_instruction(
    circuit: stim.Circuit, added_circuit: stim.Circuit
) -> stim.Circuit:
    if circuit[-1].name == "TICK":
        return circuit[:-1] + added_circuit + _tick_circuit()
    else:
        return circuit + added_circuit


def compile_fragments_to_circuit_with_detectors(
    fragments: list[Fragment | FragmentLoop],
    qubit_coords_map: dict[int, tuple[float, ...]],
) -> stim.Circuit:
    flows = build_flows_from_fragments(fragments)
    detectors_from_flows = match_detectors_from_flows_shallow(flows, qubit_coords_map)

    circuit = stim.Circuit()
    number_of_spatial_coordinates = len(
        next(iter(qubit_coords_map.values()), cast(tuple[float, ...], tuple()))
    )
    for fragment, detectors in zip(fragments, detectors_from_flows):
        detectors_circuit = _detectors_to_circuit(detectors, [0.0])
        if isinstance(fragment, Fragment):
            circuit += _insert_before_last_tick_instruction(
                fragment.circuit, detectors_circuit
            )
        else:  # isinstance(fragment, FragmentLoop):
            shift_circuit = _shift_time_instruction(number_of_spatial_coordinates)
            loop_body = compile_fragments_to_circuit_with_detectors(
                fragment.fragments, qubit_coords_map
            )
            circuit += (
                _insert_before_last_tick_instruction(
                    loop_body, shift_circuit + detectors_circuit
                )
                * fragment.repetitions
            )
    return remove_duplicate_detectors(circuit)



from __future__ import annotations

from pathlib import Path

import pytest
import stim

from tqecd.construction import annotate_detectors_automatically
from tqecd.exceptions import TQECDException
from tqecd.utils import (
    detector_to_targets_tuple,
    push_all_detectors_to_the_end,
    remove_annotations,
)

_HERE = Path(__file__).parent
_TEST_FOLDER = _HERE / "test_files"
_VALID_TEST_FOLDER = _TEST_FOLDER / "valid"
_INVALID_TEST_FOLDER = _TEST_FOLDER / "invalid"


def valid_test_circuits() -> list[tuple[str, stim.Circuit]]:
    valid_circuits: list[tuple[str, stim.Circuit]] = []
    for filepath in _VALID_TEST_FOLDER.iterdir():
        with open(filepath) as f:
            valid_circuits.append(
                (
                    filepath.name,
                    push_all_detectors_to_the_end(stim.Circuit(f.read())),
                )
            )
    return valid_circuits


def parse_invalid_circuit(text: str) -> tuple[stim.Circuit, str]:
    expected_error_message_regex_lines: list[str] = []
    circuit_str = ""
    for line in text.splitlines():
        if line.startswith("# ") and circuit_str == "":
            expected_error_message_regex_lines.append(line[2:])
        else:
            circuit_str += line + "\n"
    return stim.Circuit(circuit_str), "\n".join(expected_error_message_regex_lines)


def invalid_test_circuits() -> list[tuple[str, stim.Circuit, str]]:
    invalid_circuits: list[tuple[str, stim.Circuit, str]] = []
    for filepath in _INVALID_TEST_FOLDER.iterdir():
        with open(filepath) as f:
            file_content = f.read()
        circuit, expected_error_message_regex = parse_invalid_circuit(file_content)
        invalid_circuits.append((filepath.name, circuit, expected_error_message_regex))
    return invalid_circuits


def get_detectors_tuples_shallow(circuit: stim.Circuit) -> list[tuple[int, ...]]:
    detectors_tuples: list[tuple[int, ...]] = []
    for inst in circuit:
        if inst.name == "DETECTOR":
            assert isinstance(inst, stim.CircuitInstruction)
            detectors_tuples.append(detector_to_targets_tuple(inst))
    return detectors_tuples


@pytest.mark.parametrize("name,circuit", valid_test_circuits())
def test_valid_circuits(name: str, circuit: stim.Circuit) -> None:
    circuit_without_detectors = remove_annotations(
        circuit, frozenset(["DETECTOR", "SHIFT_COORDS"])
    )
    annotated_circuit = push_all_detectors_to_the_end(
        annotate_detectors_automatically(circuit_without_detectors)
    )

    initial_detectors = set(get_detectors_tuples_shallow(circuit))
    computed_detectors = set(get_detectors_tuples_shallow(annotated_circuit))

    missing_detectors = initial_detectors.difference(computed_detectors)
    assert not missing_detectors, "Detectors in original circuit are missing."


@pytest.mark.parametrize("name,circuit,error_message", invalid_test_circuits())
def test_invalid_circuits(name: str, circuit: stim.Circuit, error_message: str) -> None:
    circuit_without_detectors = remove_annotations(
        circuit, frozenset(["DETECTOR", "SHIFT_COORDS"])
    )
    with pytest.raises(TQECDException, match=rf"^{error_message}$"):
        annotate_detectors_automatically(circuit_without_detectors)



from __future__ import annotations

import typing as ty

import numpy
import stim

from tqecd.exceptions import TQECDException
from tqecd.pauli import PauliString

ANNOTATIONS = {
    "DETECTOR",
    "MPAD",
    "OBSERVABLE_INCLUDE",
    "QUBIT_COORDS",
    "SHIFT_COORDS",
    "TICK",
}


def iter_stim_circuit_by_moments(
    circuit: stim.Circuit,
) -> ty.Iterator[stim.Circuit | stim.CircuitRepeatBlock]:
    """Iterate over the `stim.Circuit` by moments.

    A moment in a `stim.Circuit` is a sequence of instructions between two `TICK`
    instructions. Note that we always consider a `stim.CircuitRepeatBlock` as a
    single moment.

    Args:
        circuit: The circuit to iterate over.

    Yields:
        A `stim.Circuit` or a `stim.CircuitRepeatBlock` instance.

    Invariants:
        - All the instructions of the provided circuit (even TICK ones) are
          eventually yielded by the generator returned by this function.
        - In the yield items, TICK instructions can only appear at the end of a `stim.Circuit`
          or within the body of a `stim.CircuitRepeatBlock`.
    """
    cur_moment = stim.Circuit()
    for inst in circuit:
        if isinstance(inst, stim.CircuitRepeatBlock):
            if cur_moment:
                yield cur_moment
                cur_moment.clear()
            yield inst
        elif inst.name == "TICK":
            cur_moment.append(inst)
            yield cur_moment
            cur_moment.clear()
        else:
            cur_moment.append(inst)
    if cur_moment:
        yield cur_moment


def is_measurement(instruction: stim.CircuitInstruction) -> bool:
    return instruction.name in ["M", "MR", "MRX", "MRY", "MRZ", "MX", "MY", "MZ"]


def is_reset(instruction: stim.CircuitInstruction) -> bool:
    return stim.gate_data(instruction.name).is_reset  # type: ignore


def is_noisy_gate(instruction: stim.CircuitInstruction) -> bool:
    return (
        not is_measurement(instruction)
        and stim.gate_data(instruction.name).is_noisy_gate
    )


def is_annotation(instruction: stim.CircuitInstruction) -> bool:
    return instruction.name in ANNOTATIONS


def is_virtual_instruction(inst: stim.CircuitInstruction) -> bool:
    return is_annotation(inst) or is_noisy_gate(inst)


def is_combined_measurement_reset(instruction: stim.CircuitInstruction) -> bool:
    return is_measurement(instruction) and is_reset(instruction)


def has_combined_measurement_reset(moment: stim.Circuit) -> bool:
    """Check if a `stim.Circuit` moment contains combined instructions.

    Combined instructions are instructions that implement both a measurement and
    a reset in one step.

    Args:
        moment: The moment to check.

    Returns:
        `True` if the provided moment has a combined instruction, else `False`.
    """
    for inst in moment:
        if is_virtual_instruction(inst):  # type: ignore
            continue
        if is_combined_measurement_reset(inst):  # type: ignore
            return True
    return False


def has_circuit_repeat_block(moment: stim.Circuit) -> bool:
    """Check if a `stim.Circuit` moment contains a `stim.CircuitRepeatBlock`.

    Args:
        moment: The moment to check.

    Returns:
        `True` if the provided moment has a `stim.CircuitRepeatBlock`, else `False`.
    """
    return any(isinstance(inst, stim.CircuitRepeatBlock) for inst in moment)


def has_measurement(moment: stim.Circuit) -> bool:
    """Check if a `stim.Circuit` moment contains measurement instructions.

    Args:
        moment: The moment to check.

    Returns:
        `True` if the provided moment has a measurement, else `False`.
    """
    return any(is_measurement(inst) for inst in moment)  # type:ignore


def has_only_measurement_or_is_virtual(moment: stim.Circuit) -> bool:
    """Check if a `stim.Circuit` moment contains only measurement instructions
    or is a virtual moment.

    Note:
        Annotations are ignored by this function.

    Args:
        moment: The moment to check.

    Returns:
        `True` if the provided moment has a measurement, else `False`.
        Note that this function returns `True` on a moment that only contains
        annotations and noisy-gates (measurements excluded).
    """
    for inst in moment:
        if is_virtual_instruction(inst):  # type: ignore
            continue
        if not is_measurement(inst):  # type: ignore
            return False
    return True


def has_reset(moment: stim.Circuit) -> bool:
    """Check if a `stim.Circuit` moment contains reset instructions.

    Args:
        moment: The moment to check.

    Returns:
        `True` if the provided moment has a reset, else `False`.
    """
    return any(is_reset(inst) for inst in moment)  # type:ignore


def has_only_reset_or_is_virtual(moment: stim.Circuit) -> bool:
    """Check if a `stim.Circuit` moment contains only reset instructions or is
    a virtual moment.

    Note:
        Annotations are ignored by this function.

    Args:
        moment: The moment to check.

    Returns:
        `True` if the provided moment has only reset instructions, else `False`.
        Note that this function returns `True` on a moment that only contains
        annotations and noisy-gates (measurements excluded).
    """
    for inst in moment:
        if is_virtual_instruction(inst):  # type: ignore
            continue
        if not is_reset(inst):  # type: ignore
            return False
    return True


def is_virtual_moment(moment: stim.Circuit) -> bool:
    """Check if the provided moment only contains virtual instructions.

    Virtual instructions are noisy instructions or instructions whose name is
    in the `ANNOTATIONS` structure.

    Args:
        moment: circuit to check.

    Returns:
        `True` if the provided circuit is only composed of virtual instructions,
        else `False`.
    """
    if any(isinstance(inst, stim.CircuitRepeatBlock) for inst in moment):
        raise TQECDException(
            "Breaking invariant: you provided a circuit with a stim.CircuitRepeatBlock "
            "instruction to is_virtual_moment. This is not supported."
        )
    return all(is_virtual_instruction(inst) for inst in moment)  # type:ignore


def has_computation_instruction(moment: stim.Circuit) -> bool:
    return any(
        not is_virtual_instruction(inst)  # type: ignore
        and not is_reset(inst)  # type: ignore
        and not is_measurement(inst)  # type: ignore
        for inst in moment
    )


def is_collapsing_instruction(instruction: stim.CircuitInstruction) -> bool:
    return is_reset(instruction) or is_measurement(instruction)


def pauli_string_mean_coords(
    pauli_string: PauliString, qubit_coords_map: dict[int, list[float]]
) -> tuple[float, ...]:
    all_coords_items = [qubit_coords_map[i] for i in pauli_string.qubits]
    return tuple(numpy.mean(numpy.asarray(all_coords_items), axis=0)) + (0.0,)


def _collapsing_inst_to_pauli_strings(
    inst: stim.CircuitInstruction,
) -> list[PauliString]:
    """Create the `PauliString` instances representing the provided collapsing
    instruction.

    Args:
        inst: a collapsing instruction.

    Raises:
        TQECDException: If the provided collapsing instruction has any non-qubit target.
        TQECDException: If the provided instruction is not a collapsing instruction.

    Returns:
        a list of `PauliString` instances representing the collapsing instruction
        provided as input.
    """
    name = inst.name
    targets = inst.targets_copy()
    if any(not t.is_qubit_target for t in targets):
        raise TQECDException(
            "Found a stim instruction with non-qubit target. This is not supported."
        )
    if name in ["RX", "MX", "MRX"]:
        return [PauliString({target.qubit_value: "X"}) for target in targets]  # type: ignore
    if name in ["RY", "MY", "MRY"]:
        return [PauliString({target.qubit_value: "Y"}) for target in targets]  # type: ignore
    if name in ["R", "RZ", "M", "MZ", "MR", "MRZ"]:
        return [PauliString({target.qubit_value: "Z"}) for target in targets]  # type: ignore
    raise TQECDException(
        f"Not a supported collapsing instruction: {name}. "
        "See https://github.com/quantumlib/Stim/wiki/Stim-v1.13-Gate-Reference "
        "for a list of collapsing instructions."
    )


def collapse_pauli_strings_at_moment(moment: stim.Circuit) -> list[PauliString]:
    """Compute and return the list of PauliString instances representing all
    the collapsing operations found in the provided moment.

    This function has the following pre-condition: all the instructions in the provided
    moment should be instances of `stim.CircuitInstruction`.

    This pre-condition can be ensured by only providing `stim.Circuit` instances returned
    by the `iter_stim_circuit_by_moments` function, ensuring before calling that the
    moment is not an instance of `stim.CircuitRepeatBlock`.

    This function ensures that the returned Pauli strings are ordered w.r.t the
    qubit index the collapsing operation was applied on.

    Args:
        moment: A circuit moment that does not contain any `stim.CircuitRepeatBlock`
            instance.

    Raises:
        TQECDException: If the pre-conditions of this function are not met.

    Returns:
        instances of `PauliString` representing each collapsing operation found in the
            provided moment, sorted by qubits.
    """
    # Pre-condition check
    if any(isinstance(inst, stim.CircuitRepeatBlock) for inst in moment):
        raise TQECDException(
            "Breaking pre-condition: collapse_pauli_strings_at_moment is expecting "
            f"moments without repeat blocks. Found:\n{moment}"
        )

    pauli_strings: list[PauliString] = []
    for inst in moment:
        assert isinstance(inst, stim.CircuitInstruction)
        if not is_collapsing_instruction(inst):
            continue
        collapsing_operations = _collapsing_inst_to_pauli_strings(inst)
        pauli_strings.extend(collapsing_operations)
    return pauli_strings


def remove_annotations(
    circuit: stim.Circuit,
    annotations_to_remove: frozenset[str] = frozenset(
        {"DETECTOR", "MPAD", "OBSERVABLE_INCLUDE", "SHIFT_COORDS"}
    ),
) -> stim.Circuit:
    """Remove all the annotations from a given circuit, except TICK
    instructions.

    Args:
        circuit: the circuit to remove annotations from.

    Returns:
        a new quantum circuit that contains all the instructions from the provided
        circuit except annotations that are not TICK instructions.`
    """
    new_circuit = stim.Circuit()
    for inst in circuit:
        if isinstance(inst, stim.CircuitRepeatBlock):
            new_circuit.append(
                stim.CircuitRepeatBlock(
                    repeat_count=inst.repeat_count,
                    body=remove_annotations(inst.body_copy(), annotations_to_remove),
                )
            )
        elif not is_annotation(inst) or inst.name not in annotations_to_remove:
            new_circuit.append(inst)
    return new_circuit


def _offset_detectors(
    detectors: list[stim.CircuitInstruction], offset: int
) -> list[stim.CircuitInstruction]:
    offset_detectors: list[stim.CircuitInstruction] = []
    for detector in detectors:
        targets = [stim.target_rec(t.value - offset) for t in detector.targets_copy()]
        offset_detectors.append(
            stim.CircuitInstruction("DETECTOR", targets, detector.gate_args_copy())
        )
    return offset_detectors


def detector_to_targets_tuple(instruction: stim.CircuitInstruction) -> tuple[int, ...]:
    return tuple(t.value for t in instruction.targets_copy())


def push_all_detectors_to_the_end(circuit: stim.Circuit) -> stim.Circuit:
    new_circuit = stim.Circuit()
    detectors: list[stim.CircuitInstruction] = []
    for instruction in circuit:
        if isinstance(instruction, stim.CircuitRepeatBlock):
            new_circuit.append(
                stim.CircuitRepeatBlock(
                    instruction.repeat_count,
                    push_all_detectors_to_the_end(instruction.body_copy()),
                )
            )
        elif is_measurement(instruction):
            num_targets = len(instruction.targets_copy())
            detectors = _offset_detectors(detectors, num_targets)
            new_circuit.append(instruction)
        elif instruction.name == "DETECTOR":
            detectors.append(instruction)
        else:
            new_circuit.append(instruction)

    for detector in sorted(detectors, key=detector_to_targets_tuple, reverse=True):
        new_circuit.append(detector)
    return new_circuit


def remove_duplicate_detectors(circuit: stim.Circuit) -> stim.Circuit:
    """Remove duplicate detectors from the provided circuit.

    Detectors are considered duplicates if they have the same measurement record
    targets even if the coordinates are different.

    Here we simplify the measurements tracking by always assuming the detectors
    are still valid even if the repeatition count for the ``stim.CircuitRepeatBlock``
    instructions equals one.

    Args:
        circuit: circuit that includes detector instructions to remove duplicates from.

    Returns:
        A new `stim.Circuit` instance with duplicate detectors removed.
    """
    seen_detectors: set[frozenset[int]] = set()
    num_measurements: int = 0

    def remove_impl(circ: stim.Circuit) -> stim.Circuit:
        nonlocal seen_detectors, num_measurements
        new_circuit = stim.Circuit()
        for inst in circ:
            if isinstance(inst, stim.CircuitRepeatBlock):
                new_circuit.append(
                    stim.CircuitRepeatBlock(
                        repeat_count=inst.repeat_count,
                        body=remove_impl(inst.body_copy()),
                    )
                )
            elif inst.name == "DETECTOR":
                targets = frozenset(
                    t.value + num_measurements for t in inst.targets_copy()
                )
                if targets in seen_detectors:
                    continue
                seen_detectors.add(targets)
                new_circuit.append(inst)
            elif is_measurement(inst):
                num_measurements += len(inst.targets_copy())
                new_circuit.append(inst)
            else:
                new_circuit.append(inst)
        return new_circuit

    return remove_impl(circuit)



"""Provides a few utility functions to match flows.

This module provides functions to encode the problem of finding matching flows
as a SAT problem, as well as a function to solve that SAT problem.

The three functions re-exported by this module all aims at finding a list of
Pauli strings from a given list that "cover" a target Pauli string:

- :func:`~.cover.find_cover` performs a brute-force search to find Pauli strings
  from a provided candidate list that, once multiplied together, exactly match
  the provided target Pauli string,
- :func:`~.cover.find_exact_cover_sat` performs the same task as
  :func:`~.cover.find_cover` but encodes the problem as a SAT problem and uses a
  SAT solver to find potential solutions instead of performing a brute-force
  search.
- :func:`~.cover.find_commuting_cover_on_target_qubits_sat` uses the same
  strategy as :func:`~.cover.find_exact_cover_sat` by using a SAT solver, but
  solves a slightly different problem that consists in finding Pauli strings
  that, once multiplied together, commute with a provided target Pauli string
  (instead of being exactly equal to it).
"""

from .cover import (
    find_commuting_cover_on_target_qubits_sat as find_commuting_cover_on_target_qubits_sat,
)
from .cover import find_cover as find_cover
from .cover import find_exact_cover_sat as find_exact_cover_sat



from __future__ import annotations

from dataclasses import dataclass

from tqecd.exceptions import TQECDException


@dataclass(frozen=True)
class RelativeMeasurementLocation:
    """Represents a unique measurement with an offset.

    Measurements are represented with the same method as stim: by a negative
    offset. This means that instances of this class have no meaning if they
    are not associated with a precise point in the circuit, but this also
    means that they can represent measurements without the need for global
    information.

    This class also stores the qubit that is being measured in order to be able
    to retrieve its coordinate to build a detector when found.
    """

    offset: int
    qubit_index: int

    def __post_init__(self) -> None:
        if self.offset >= 0:
            raise TQECDException(
                "Relative measurement offsets should be strictly negative."
            )

    def offset_by(self, offset: int) -> RelativeMeasurementLocation:
        return RelativeMeasurementLocation(
            offset=self.offset + offset, qubit_index=self.qubit_index
        )


def get_relative_measurement_index(
    all_measured_qubits: list[int], measured_qubit: int
) -> RelativeMeasurementLocation:
    return RelativeMeasurementLocation(
        all_measured_qubits.index(measured_qubit) - len(all_measured_qubits),
        measured_qubit,
    )



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from typing import Final, Iterator, Mapping

import numpy
import stim

from tqecd.boundary import BoundaryStabilizer
from tqecd.exceptions import TQECDException
from tqecd.flow import FragmentFlows, FragmentLoopFlows
from tqecd.match_utils.cover import find_exact_cover_sat
from tqecd.measurement import RelativeMeasurementLocation


@dataclass(frozen=True)
class MatchedDetector:
    """Represents an automatically computed detector."""

    coords: tuple[float, ...]
    measurements: frozenset[RelativeMeasurementLocation]
    resets: tuple[frozenset[int], ...]

    def __hash__(self) -> int:
        return hash(self.measurements)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MatchedDetector):
            return False
        if self.measurements != other.measurements:
            return False
        return numpy.allclose(self.coords, other.coords)

    def to_instruction(self) -> stim.CircuitInstruction:
        targets = [
            stim.target_rec(m.offset)
            for m in sorted(
                self.measurements, key=lambda measurement: -measurement.offset
            )
        ]
        return stim.CircuitInstruction("DETECTOR", targets, list(self.coords))

    def with_time_coordinate(self, time_coordinate: float) -> MatchedDetector:
        return MatchedDetector(
            self.coords + (time_coordinate,), self.measurements, self.resets
        )


def _get_detectors_with_time_coordinate(
    flows: list[FragmentFlows | FragmentLoopFlows],
    matched_detectors: list[list[MatchedDetector]],
) -> list[list[MatchedDetector]]:
    updated_detectors: list[list[MatchedDetector]] = []
    number_of_fragment_flows_seen = 0
    for flow, detectors in zip(flows, matched_detectors):
        if isinstance(flow, FragmentFlows):
            time_coordinate = number_of_fragment_flows_seen
            number_of_fragment_flows_seen += 1
        else:
            # We insert detectors at the end of the last flow, so we need to compute
            # the number of FragmentFlows instances before it.
            time_coordinate = (
                sum(isinstance(f, FragmentFlows) for f in flow.fragment_flows) - 1
            )
        updated_detectors.append(
            [d.with_time_coordinate(time_coordinate) for d in detectors]
        )
    return updated_detectors


def match_detectors_from_flows_shallow(
    flows: list[FragmentFlows | FragmentLoopFlows],
    qubit_coordinates: dict[int, tuple[float, ...]],
) -> list[list[MatchedDetector]]:
    """Match detectors in the provided fragments.

    This function only performs a shallow search. Read attentively the warning and
    return sections of this documentation to understand how to use it correctly.

    Warning:
        This function only performs a shallow computation! It will consider REPEAT
        constructs as an atomic fragment with outgoing stabilizers on both sides,
        but will not recurse in the loop.

        In practice, this limitation should not be a huge issue. As the loops are
        still considered as an atomic operation, resets in the last fragment of
        the loop body and measurements in the first fragment of the loop body are
        still accounted for.
        This means that if the loop body only implements a simple QEC round (e.g.,
        surface code or repetition code), or more generally if the loop body is
        considered as a single :class:`Fragment` instance, the detectors will be
        appropriately matched and returned.

    Args:
        flows: a list of fragment flow to search detectors in.
        qubit_coordinates: a mapping from qubit indices to coordinates. Used to annotate
            the matched detectors with the coordinates from the qubits involved in the
            measurement forming the detector.

    Returns:
        the list of all the detectors found. These detectors are only valid if inserted
        at the end of the fragment represented by the last provided flow in the input
        parameter `flows`. If the last provided flow is an instance of
        :class:`FragmentLoopFlows` then the returned detectors should be inserted at
        the end of the loop body.
    """
    detectors: list[list[MatchedDetector]] = [
        match_detectors_within_fragment(flow, qubit_coordinates) for flow in flows
    ]
    for i in range(1, len(flows)):
        detectors[i].extend(
            match_boundary_stabilizers(flows[i - 1], flows[i], qubit_coordinates)
        )

    return _get_detectors_with_time_coordinate(flows, detectors)


def match_detectors_within_fragment(
    flows: FragmentFlows | FragmentLoopFlows,
    qubit_coordinates: Mapping[int, tuple[float, ...]],
) -> list[MatchedDetector]:
    """Match all the detectors that can be resolved without considering
    neighbouring fragments.

    Some detectors only include measurements from one fragment. This is for example the
    case during the initialisation of any code: some measurements from the first round
    have a pre-defined value that should be constant in the absence of errors.

    Note that this function ignores "trivial" flows to avoid trivial detectors to be
    matched on some particular cases (e.g., 1-round repetition code).

    Args:
        flows: pre-computed flows for the fragment of interest. If any detector
            is found, the involved flow(s) will be removed from the provided instance.
            This means that **this parameter may be mutated inline**.
        qubit_coordinates: a mapping from qubit indices to coordinates. Used to annotate
            the matched detectors with the coordinates from the qubits involved in the
            measurement forming the detector.

    Returns:
        the list of all the detectors found.
    """
    matched_detectors: list[MatchedDetector] = []
    matched_detectors.extend(
        _match_non_propagating_non_trivial_flows_inline(
            flows.creation, qubit_coordinates
        )
    )
    matched_detectors.extend(
        _match_non_propagating_non_trivial_flows_inline(
            flows.destruction, qubit_coordinates
        )
    )
    return matched_detectors


def _match_non_propagating_non_trivial_flows_inline(
    boundary_stabilizers: list[BoundaryStabilizer],
    qubit_coordinates: Mapping[int, tuple[float, ...]],
) -> list[MatchedDetector]:
    """Match all the detectors that can be trivially resolved and remove the
    matched boundary stabilizers from the provided list.

    This function try to find a trivial match for each of the provided boundary
    stabilizers. A trivial match is defined here as a flow that is fully
    collapsed by the collapsing operations it encounters, stopping its
    propagation.

    Note that an approximation is made in this function: trivial flows (not to
    be confused with trivial matches) are excluded from the search because they
    generate unwanted detectors.

    This function **mutates its input in-place**.

    This function guarantees the following:

    ```py
    boundary_stabilizers = []   # Fill it!
    qubit_coordinates = {}      # Fill it!
    len_bs_before = len(boundary_stabilizers)

    detectors = _match_non_propagating_non_trivial_flows_inline(
        boundary_stabilizers, qubit_coordinates
    )
    assert len_bs_before == len(detectors) + len(boundary_stabilizers)
    ```

    Args:
        boundary_stabilizers: list of boundary stabilizers to match trivially.
            This list is modified in-place: when a detector is found, the
            boundary_stabilizer associated with this detector is removed from
            the list.
        qubit_coordinates: a mapping from qubit indices to coordinates. Used to annotate
            the matched detectors with the coordinates from the qubits involved in the
            measurement forming the detector.

    Returns:
        all matched detectors.
    """
    matched_detectors: list[MatchedDetector] = []
    non_propagating_flows_indices = list(
        _find_non_propagating_non_trivial_flows(boundary_stabilizers)
    )

    # Note that the call to sorted here is not really needed as
    # _find_non_propagating_non_trivial_flows guarantees that the returned indices
    # are sorted, so `non_propagating_flows_indices[::-1]` should be enough.
    # We still call sorted here to avoid subtle bugs in case the above-mentioned
    # post-condition is changed one day, and because the number of indices
    # is not expected to be large here.
    for i in sorted(non_propagating_flows_indices, reverse=True):
        flow = boundary_stabilizers.pop(i)
        matched_detectors.append(
            MatchedDetector(
                coords=flow.coordinates(qubit_coordinates),
                measurements=frozenset(flow.measurements),
                resets=(flow.resets_qubits,),
            )
        )

    return matched_detectors


def _find_non_propagating_non_trivial_flows(
    boundary_stabilizers: list[BoundaryStabilizer],
) -> Iterator[int]:
    """Find all the boundary stabilizers that do not have any anti-commuting
    collapsing operations and that are fully collapsed by the commuting
    operations.

    Args:
        boundary_stabilizers: the boundary stabilizers to search in.

    Yields:
        indices (in ascending order) of boundary stabilizers in the provided
        `boundary_stabilizers` list that are non-trivial and fully collapsed
        by the commuting operations.
    """
    for i, stabilizer in enumerate(boundary_stabilizers):
        if (
            not stabilizer.is_trivial()
            and not stabilizer.has_anticommuting_operations
            and stabilizer.after_collapse.non_trivial_pauli_count == 0
        ):
            yield i


def match_boundary_stabilizers(
    left_flows: FragmentFlows | FragmentLoopFlows,
    right_flows: FragmentFlows | FragmentLoopFlows,
    qubit_coordinates: Mapping[int, tuple[float, ...]],
    perform_sanity_check: bool = True,
) -> list[MatchedDetector]:
    """Match detectors using the boundary stabilizers between the two given
    flows.

    This function try to match detectors that will be valid at the end of
    `right_flows` by looking at destruction boundary stabilizers back-propagated
    from measurements in `right_flows` and creation boundary stabilizers
    propagated from resets in `left_flows`.
    The above 2 stabilizer types should meet at the interface between the two
    provided flows. If we find any creation stabilizer matching (i.e. being
    equal to) any other destruction stabilizer, these form a detector!

    This function **mutates `left_flows` and `right_flows` in-place** if
    a detector is found.

    Args:
        left_flows: pre-computed flows from the left fragment.
        right_flows: pre-computed flows from the right fragment.
        qubit_coordinates: _description_
        perform_sanity_check: If True and if `right_flows` is an instance of
            :class:`FragmentLoopFlows`, detectors found between `left_flows`
            and `right_flows` are compared to the detectors found between the
            last and first fragments of the body of `right_flows`. If the two
            sets are not equal, an exception is raised. Defaults to True.

    Raises:
        TQECDException: if the sanity check does not pass.

    Returns:
        all matched detectors.
    """
    matched_detectors: list[MatchedDetector] = []

    # Sanity check: if the provided right flow is a FragmentLoopFlows, this function
    #     will compute the detectors between the first iteration of the loop and the
    #     previous fragment. If the circuit is valid, the computed detectors should
    #     be exactly the same between the fragments of two loop bodies (if the loop
    #     is repeated at least twice).
    #     Because the matching functions are modifying flows in-place, we pre-compute the
    #     detectors here on copies and will compare them before returning.
    should_sanity_check: Final[bool] = perform_sanity_check and (
        isinstance(right_flows, FragmentLoopFlows) and right_flows.repeat > 1
    )
    if should_sanity_check:
        matched_detectors_within_loop = match_boundary_stabilizers(
            # Type checking is disabled below. right_flows is guaranteed to be of type
            # FragmentLoopFlows (per the value of should_sanity_check), and so have a
            # "fragment_flows" attribute.
            deepcopy(right_flows.fragment_flows[-1]),  # type: ignore
            deepcopy(right_flows.fragment_flows[0]),  # type: ignore
            qubit_coordinates,
        )

    # 0. Combining anti-commuting stabilizers
    # We try to combine stabilizers anti-commuting with their collapsing operations
    # here because:
    #   1. the `try_merge_anticommuting_flows` method does not consider commuting
    #      stabilizers, so calling this method has the same cost now and after a first
    #      round of matching.
    #   2. the different matching methods are now quite efficient and so should not
    #      suffer too much (in terms of performance) from the commuting stabilizers
    #      potentially added at this step.
    # If commuting stabilizer computation is experiencing performance issues, it might
    # be interesting to perform this step AFTER a first round of commuting stabilizer
    # matching, and repeating the matching steps after to match newly added commuting
    # stabilizers into detectors.
    left_flows.try_merge_anticommuting_flows()
    right_flows.try_merge_anticommuting_flows()

    # 1. Match stabilizers 1-to-1 without anti-commuting collapses
    matched_detectors.extend(
        _match_commute_stabilizers(left_flows, right_flows, qubit_coordinates)
    )

    # 2. Try to match remaining stabilizers without any anti-commuting collapses
    #    by trying to find covers.
    matched_detectors.extend(
        _match_by_disjoint_cover(left_flows, right_flows, qubit_coordinates)
    )
    # Perform the sanity check if needed.
    if should_sanity_check:
        # Notes about the following "type: ignore":
        # - should_sanity_check is "constant" (annotated as Final) and so cannot
        #   be changed.
        # - matched_detectors_within_loop is unconditionally bound to a value if
        #   should_sanity_check is True.
        # - if we are here, then should_sanity_check is True, and so
        #   matched_detectors_within_loop should be bound to a value.
        if set(matched_detectors_within_loop) != set(matched_detectors):  # type: ignore[reportPossiblyUnboundVariable]
            raise TQECDException(
                f"The set of detectors computed from measurements in\n{left_flows}\n"
                f"and\n{right_flows}\nis not the same as the set of detectors computed "
                "from measurements between two loop body repetitions. Is your QEC "
                "circuit valid?"
            )
    return matched_detectors


def _match_commute_stabilizers(
    left_flows: FragmentFlows | FragmentLoopFlows,
    right_flows: FragmentFlows | FragmentLoopFlows,
    qubit_coordinates: Mapping[int, tuple[float, ...]],
) -> list[MatchedDetector]:
    """Take stabilizers from two contiguous fragments and try to match them.

    Let use the two fragment instances `f1` and `f2` such that `f1` is just
    before `f2`. There is a stabilizer boundary at the interface between `f1`
    and `f2`. Considering only the stabilizers that propagated forward from
    the resets of `f1` and commuted with all the measurements they encountered
    as well as the stabilizers that propagated backwards from the measurements
    of `f2` and commuted with all the resets they encountered, we can try to
    match these stabilizers at the interface.

    This method performs a "dumb" matching between stabilizers, only matching
    if one stabilizer from `f1` exactly matches with one stabilizer from `f2`.

    Args:
        left_flows: flows of the fragment that comes first.
        right_flows: flows of the fragment that comes second.
        qubit_coordinates: a mapping from qubit indices to coordinates. Used to
            annotate the matched detectors with the coordinates from the qubits
            involved in the measurement forming the detector.

    Returns:
        the list of all the detectors found.
    """
    detectors: list[MatchedDetector] = []
    left_flows_indices_to_remove: list[int] = []

    for ilhs, creation_flow in enumerate(left_flows.creation):
        if creation_flow.has_anticommuting_operations:
            continue

        right_flows_index_to_remove: int | None = None
        for irhs, destruction_flow in enumerate(right_flows.destruction):
            if destruction_flow.has_anticommuting_operations:
                continue

            # If the stabilizers are not exactly the same, this is not a match.
            if creation_flow.after_collapse != destruction_flow.after_collapse:
                continue
            # Else, it is a match!
            measurements = frozenset(
                m.offset_by(-right_flows.total_number_of_measurements)
                for m in creation_flow.measurements
            ) | frozenset(destruction_flow.measurements)
            resets = (creation_flow.resets_qubits, destruction_flow.resets_qubits)
            detectors.append(
                MatchedDetector(
                    coords=destruction_flow.coordinates(qubit_coordinates),
                    measurements=measurements,
                    resets=resets,
                )
            )
            left_flows_indices_to_remove.append(ilhs)
            right_flows_index_to_remove = irhs
            # break here because the creation stabilizer is matched and
            # will be removed, so there is no point trying to match it with
            # other destruction stabilizers.
            break
        # Here, we finished the iteration on right_flows, so we can remove
        # the matched flows.
        if right_flows_index_to_remove is not None:
            right_flows.remove_destruction(right_flows_index_to_remove)

    # Here, we finished the iteration on left_flows, so we can remove the
    # matched flows.
    left_flows.remove_creations(left_flows_indices_to_remove)
    return detectors


def _match_boundary_stabilizers_by_disjoint_cover(
    target_stabilizers: list[BoundaryStabilizer],
    covering_stabilizers: list[BoundaryStabilizer],
    qubit_coordinates: Mapping[int, tuple[float, ...]],
    target_is_creation: bool,
) -> tuple[list[MatchedDetector], list[int]]:
    """Try to match stabilizers by finding a cover.

    This function will try to match each of the stabilizers found in
    `target_stabilizers` with several stabilizers from `covering_stabilizers`.
    This is a one-to-many relationship: one stabilizer from `target_stabilizers`
    will be matched to many stabilizers from `covering_stabilizers`.

    Args:
        target_stabilizers: list of stabilizers to try to separately match with
            several stabilizers from `covering_stabilizers`.
        covering_stabilizers: list of stabilizers that can be combined together to
            match with one stabilizer from `target_stabilizers`.
        qubit_coordinates: a mapping from qubit indices to coordinates. Used to
            annotate the matched detectors with the coordinates from the qubits
            involved in the measurement forming the detector.
        target_is_creation: `True` if the provided `target_stabilizers` are
            creation stabilizers, else `False`.
    Returns:
        the list of all the detectors found as well as the indices of stabilizers
        from `target_stabilizers` that have been used and should be removed from
        the considered stabilizers.
    """
    detectors: list[MatchedDetector] = []
    targets_to_remove: list[int] = []
    # For each target, try to find a cover.
    for i, target in enumerate(target_stabilizers):
        # Try to find the cover. If unsuccessful, skip `target` and continue the loop.
        # Else, a detector is matched, so insert it into the returned list and mark
        # `target` as "to remove".
        cover_indices = find_exact_cover_sat(
            target.after_collapse, [cs.after_collapse for cs in covering_stabilizers]
        )
        if cover_indices is None:
            continue

        measurements_involved_in_cover: frozenset[RelativeMeasurementLocation] = (
            frozenset()
        )
        resets_qubits_involved_in_cover: frozenset[int] = frozenset()
        for j in cover_indices:
            # We know for sure that each boundary stabilizer commutes with the collapsing
            # operations on the boundary. That means that if a measurement is appearing
            # an even number of times, the associated flow would simply cancel out, not
            # touching the measurement. The symmetric difference operator is used for that
            # purpose here.
            measurements_involved_in_cover ^= set(covering_stabilizers[j].measurements)
            resets_qubits_involved_in_cover ^= covering_stabilizers[j].resets_qubits

        measurements = frozenset(target.measurements) | measurements_involved_in_cover
        resets = (
            (target.resets_qubits, resets_qubits_involved_in_cover)
            if target_is_creation
            else (resets_qubits_involved_in_cover, target.resets_qubits)
        )
        detectors.append(
            MatchedDetector(
                coords=target.coordinates(qubit_coordinates),
                measurements=measurements,
                resets=resets,
            )
        )
        targets_to_remove.append(i)
    return detectors, targets_to_remove


def _match_by_disjoint_cover(
    left_flows: FragmentFlows | FragmentLoopFlows,
    right_flows: FragmentFlows | FragmentLoopFlows,
    qubit_coordinates: Mapping[int, tuple[float, ...]],
) -> list[MatchedDetector]:
    """Try to match stabilizers by finding a cover.

    Args:
        left_flows: flows of the fragment that comes first.
        right_flows: flows of the fragment that comes second.
        qubit_coordinates: a mapping from qubit indices to coordinates. Used to
            annotate the matched detectors with the coordinates from the qubits
            involved in the measurement forming the detector.

    Returns:
        the list of all the detectors found.
    """
    # Build the correct stabilizer lists
    left_boundary_stabilizers: list[BoundaryStabilizer] = []
    left_boundary_indices_map: list[int] = []
    for i, stabilizer in enumerate(left_flows.creation):
        if not stabilizer.has_anticommuting_operations:
            left_boundary_stabilizers.append(
                stabilizer.with_measurement_offset(
                    -right_flows.total_number_of_measurements
                )
            )
            left_boundary_indices_map.append(i)

    right_boundary_stabilizers: list[BoundaryStabilizer] = []
    right_boundary_indices_map: list[int] = []
    for i, stabilizer in enumerate(right_flows.destruction):
        if not stabilizer.has_anticommuting_operations:
            right_boundary_stabilizers.append(stabilizer)
            right_boundary_indices_map.append(i)

    # Check that there is at least a hope that a cover exist
    if len(left_boundary_stabilizers) == 0 or len(right_boundary_stabilizers) == 0:
        return []
    if len(left_boundary_stabilizers) == 1 and len(right_boundary_stabilizers) == 1:
        return []

    # Try to find a cover composed of measurements from the right flows that will
    # cancel the stabilizer propagated from one reset in the left flows.
    forward_detectors, left_reduced_indices_to_remove = (
        _match_boundary_stabilizers_by_disjoint_cover(
            left_boundary_stabilizers,
            right_boundary_stabilizers,
            qubit_coordinates,
            target_is_creation=True,
        )
    )
    # Remove all the left flows that have been matched with a cover. Both the actual flow
    # and the temporary list created above should be updated.
    for i in sorted(left_reduced_indices_to_remove, reverse=True):
        left_boundary_stabilizers.pop(i)
        left_flows.remove_creation(left_boundary_indices_map[i])

    # Try to find a cover composed of resets from the left flows that will cancel the
    # stabilizer propagated backward from one measurement in the right flows.
    backward_detectors, right_reduced_indices_to_remove = (
        _match_boundary_stabilizers_by_disjoint_cover(
            right_boundary_stabilizers,
            left_boundary_stabilizers,
            qubit_coordinates,
            target_is_creation=False,
        )
    )
    # Remove all the right flows that have been matched with a cover. Both the actual flow
    # and the temporary list created above should be updated.
    for i in sorted(right_reduced_indices_to_remove, reverse=True):
        right_boundary_stabilizers.pop(i)
        right_flows.remove_destruction(right_boundary_indices_map[i])

    # Return all the detectors that have been found.
    return forward_detectors + backward_detectors



"""Provides utility functions to find "covers" of Pauli strings."""

from __future__ import annotations

import time
from typing import Iterator

import pysat.solvers

from tqecd.boundary import BoundaryStabilizer, manhattan_distance
from tqecd.match_utils.sat import (
    encode_pauli_string_commuting_cover_sat_problem_in_solver,
    encode_pauli_string_exact_cover_sat_problem_in_solver,
)
from tqecd.pauli import PauliString


def _all_pauli_string_combination_results(
    pauli_strings: list[PauliString],
) -> Iterator[tuple[list[bool], PauliString]]:
    """Iterate over all the possible Pauli string products.

    This function iterates over all the ``2**len(pauli_strings)`` products
    that may be generated by either picking or not picking any of the
    provided Pauli strings.

    It is efficient in the sense that it performs the minimum number of
    Pauli products required to do so.

    Args:
        pauli_strings: a list of Pauli strings that will be considered
            in the returned products.

    Yields:
        all the ``2**len(pauli_strings)`` possible Pauli products, one by one.
    """
    yield from _all_pauli_string_combination_results_impl(
        pauli_strings, PauliString({})
    )


def _all_pauli_string_combination_results_impl(
    pauli_strings: list[PauliString], current_pauli_string: PauliString
) -> Iterator[tuple[list[bool], PauliString]]:
    """Iterate over all the possible Pauli string products.

    This function iterates over all the ``2**len(pauli_strings)`` products
    that may be generated by either picking or not picking any of the
    provided Pauli strings.

    It is efficient in the sense that it performs the minimum number of
    Pauli products required to do so.

    Args:
        pauli_strings: a list of Pauli strings that will be considered
            in the returned products.

    Yields:
        tuples containing a list of boolean of the exact same size as the
        provided ``pauli_strings`` list and representing the Pauli strings
        that have been considered in the product returned as the second
        entry of the tuple.
    """
    if len(pauli_strings) == 0:
        yield [], current_pauli_string
        return
    yield from (
        ([True] + choices, res_pauli_string)
        for choices, res_pauli_string in _all_pauli_string_combination_results_impl(
            pauli_strings[1:],
            current_pauli_string * pauli_strings[0],
        )
    )
    yield from (
        ([False] + choices, res_pauli_string)
        for choices, res_pauli_string in _all_pauli_string_combination_results_impl(
            pauli_strings[1:], current_pauli_string
        )
    )


def find_cover(
    target: BoundaryStabilizer,
    sources: list[BoundaryStabilizer],
    qubit_coordinates: dict[int, tuple[float, ...]],
    maximum_qubit_distance: int = 5,
) -> list[BoundaryStabilizer] | None:
    """Try to cover the provided ``target`` stabilizer with stabilizers from
    ``sources``.

    This function is currently performing a bruteforce search: it tries all
    combinations of stabilizers from ``sources`` and check if one matches with
    the provided ``target``.
    This approach blows up very quickly, as it may test up to ``2**len(sources)``
    different combinations before being able to tell that no match has been found.

    Note that **all** the combinations are tested, even the only combination that
    consist in "picking nothing from the list", resulting in an empty PauliString.
    To avoid any surprise, and because we do not expect an empty ``target`` to make
    sense here, this function will raise on such a case.

    Args:
        target: the boundary stabilizer to try to match with the provided
            ``sources``.
        sources: the boundary stabilizers that this function will try to combine to
            find a stabilizer involving ``target``.
        qubit_coordinates: a mapping from qubit indices to coordinates. Used to
            annotate the matched detectors with the coordinates from the qubits
            involved in the measurement forming the detector.
        maximum_qubit_distance: radius (in number of qubits) to consider when
            searching for covering boundary stabilizers. Any boundary stabilizer
            in ``sources`` that has coordinates outside of that radius from
            ``target`` will not be considered, reducing the overall complexity
            of this function. The radius is computed with the Manhattan distance.

    Raises:
        TQECDException: if ``target`` is the empty PauliString.
        TQECDException: if any of the provided instances in ``target`` or
            ``sources`` has anti-commuting stabilizers.

    Returns:
        a matching set of boundary stabilizers or None if no matching stabilizers
        could be found.
    """

    sources = [
        s
        for s in sources
        if manhattan_distance(target, s, qubit_coordinates) <= maximum_qubit_distance
    ]

    after_collapse_sources = [s.after_collapse for s in sources]
    for (
        picked_stabilizers,
        resulting_pauli_string,
    ) in _all_pauli_string_combination_results(after_collapse_sources):
        if target.after_collapse == resulting_pauli_string:
            return [
                boundary_stabilizer
                for i, boundary_stabilizer in enumerate(sources)
                if picked_stabilizers[i]
            ]
    return None


def _smallest_solution_shortcircuit(
    solutions: Iterator[list[int]], lower_length_bound: int = 0, timeout: float = 0.1
) -> list[int] | None:
    """Iterate over the provided ``solutions`` iterator to find the smallest
    possible solution within the provided ``timeout``.

    Args:
        solutions: iterator yielding lists of integers representing the indices of
            Pauli strings that should be used to cover a specific target.
        lower_length_bound: The minimum expected length. This parameter is used to
            quit early if a solution with this length is found. Default to exploring
            all the solutions to find the best one.
        timeout: maximum time in seconds that can be spent finding the smallest
            solution. The provided ``solutions`` iterator might yield millions of
            inputs, so this parameter is here to stop the search and return the best
            found solution when more than ``timeout`` seconds have been spent within
            this function.

    Returns:
        the smallest solution found, or ``None`` if no solution was found.
    """
    start_time: float = time.monotonic()
    smallest_solution = next(solutions, None)

    if smallest_solution is None:
        return None
    if len(smallest_solution) == lower_length_bound:
        return smallest_solution

    for solution in solutions:
        if len(solution) == lower_length_bound:
            return solution
        smallest_solution = min((smallest_solution, solution), key=len)
        if time.monotonic() - start_time > timeout:
            return smallest_solution
    return smallest_solution


def _find_cover_sat(
    target: PauliString, sources: list[PauliString], on_qubits: frozenset[int]
) -> list[int] | None:
    """Try to find a set of boundary stabilizers from ``sources`` that generate
    target on qubits ``on_qubits``.

    If multiple valid covers exist, the covers involving the lowest number of
    :class:`~tqecd.pauli.PauliString` instances from ``sources`` are listed, and
    a random cover is picked from that list.

    Args:
        target: the stabilizers to cover with stabilizers from ``sources``.
        sources: stabilizers that can be used to cover `target`.
        on_qubits: qubits to consider when trying to cover ``target`` with
            ``sources``.

    Returns:
        Either a list of indices over ``sources`` that, when combined, cover
        exactly the provided ``target`` on all the qubits provided in
        ``on_qubits``, or ``None`` if such a list could not be found.
    """
    with pysat.solvers.CryptoMinisat() as solver:
        encode_pauli_string_exact_cover_sat_problem_in_solver(
            solver, target, sources, on_qubits
        )
        return _smallest_solution_shortcircuit(
            (
                [i - 1 for i in satisfying_proof if i > 0]
                for satisfying_proof in solver.enum_models()
            ),
            2,
        )


def find_exact_cover_sat(
    target: PauliString, sources: list[PauliString]
) -> list[int] | None:
    """Try to find a set of pauli strings from ``sources`` that generate exactly
    ``target``.

    The Pauli strings returned (via indices over the provided ``sources``), once
    multiplied together, should be exactly equal to ``target``. In particular, the
    following post-condition should hold:

    If multiple valid covers exist, the covers involving the lowest number of
    :class:`~tqecd.pauli.PauliString` instances from ``sources`` are listed, and
    a random cover is picked from that list.

    .. code-block:: python

        target = None     # to replace
        sources = [None]  # to replace
        cover_indices = find_exact_cover_sat(target, sources)
        resulting_pauli_string = PauliString({})
        for i in cover_indices:
            resulting_pauli_string = resulting_pauli_string * sources[i]
        assert resulting_pauli_string == target, "Should hold"

    Args:
        target: the stabilizers to cover with stabilizers from ``sources``.
        sources: stabilizers that can be used to cover ``target``.

    Returns:
        Either a list of indices over ``sources`` that, when combined, cover
        exactly the provided ``target``, or ``None`` if such a list could not be
        found.
    """
    # If target is the identity, we do not have to call a SAT solver to find
    # a solution: pick no Pauli string.
    # Note: we might want to disallow an empty return in the future.
    if target.non_trivial_pauli_count == 0:
        return []

    # Else, if there are no sources, we cannot find a solution.
    if not sources:
        return None

    # We want an exact (i.e., equality) cover on all qubits, to be sure that
    # the post-condition in the docstring holds. For that, it is sufficient to
    # only consider all the qubits where either `target` or at least one of the
    # items of `sources` acts non-trivially (i.e., something else than the identity).
    involved_qubits = frozenset(target.qubits)
    for source in sources:
        involved_qubits |= frozenset(source.qubits)

    return _find_cover_sat(target, sources, involved_qubits)


def find_commuting_cover_on_target_qubits_sat(
    target: PauliString, sources: list[PauliString]
) -> list[int] | None:
    """Try to find a set of boundary stabilizers from ``sources`` that generate a
    superset of ``target``.

    This function try to find a set of Pauli strings from ``sources`` that
    includes ``target`` (i.e., on every qubit where ``target`` is non-trivial,
    the product of each of the returned Pauli strings should commute with
    ``target``).

    If multiple valid covers exist, the covers involving the lowest number of
    :class:`~tqecd.pauli.PauliString` instances from ``sources`` are listed, and
    a random cover is picked from that list.

    The differences with :func:`find_cover_sat` are:

    1. this function does not restrict the output of the product of each of
       the returned Pauli string on qubits where ``target`` acts trivially (i.e.
       "I"). So in practice, on qubits where ``target[qubit] == "I"``, the value
       of the returned Pauli string can be anything.
    2. this function does not restrict the output of the product of each of
       the returned Pauli string to exactly match ``target`` on qubits where
       it acts non-trivially, but rather requires the output to commute with
       ``target`` on those qubits.

    Args:
        target: the stabilizers to cover with stabilizers from ``sources``.
        sources: stabilizers that can be used to cover ``target``.

    Returns:
        Either a list of a stabilizers that, when combined, commute with
        the provided ``target``, or ``None`` if such a list could not be found.
    """
    if not sources:
        return None

    with pysat.solvers.CryptoMinisat() as solver:
        encode_pauli_string_commuting_cover_sat_problem_in_solver(
            solver, target, sources, frozenset(target.qubits)
        )
        return _smallest_solution_shortcircuit(
            (
                [i - 1 for i in satisfying_proof if i > 0]
                for satisfying_proof in solver.enum_models()
            ),
            2,
        )



import pytest

from tqecd.exceptions import TQECDException
from tqecd.measurement import (
    RelativeMeasurementLocation,
    get_relative_measurement_index,
)


def test_initialisation() -> None:
    RelativeMeasurementLocation(-1, 1)

    with pytest.raises(
        TQECDException,
        match=r"^Relative measurement offsets should be strictly negative\.$",
    ):
        RelativeMeasurementLocation(0, 1)
    with pytest.raises(
        TQECDException,
        match=r"^Relative measurement offsets should be strictly negative\.$",
    ):
        RelativeMeasurementLocation(1, 1)


def test_offset_by() -> None:
    rml = RelativeMeasurementLocation(-1, 1)
    assert rml.offset_by(-19) == RelativeMeasurementLocation(-20, 1)
    with pytest.raises(
        TQECDException,
        match=r"^Relative measurement offsets should be strictly negative\.$",
    ):
        rml.offset_by(1)


def test_get_relative_measurement_index() -> None:
    all_measured_qubits = list(range(10))
    for qubit in all_measured_qubits:
        rml = get_relative_measurement_index(all_measured_qubits, qubit)
        assert rml.qubit_index == qubit
        assert rml.offset == qubit - len(all_measured_qubits)



import pytest
import stim

from tqecd.exceptions import TQECDException
from tqecd.pauli import PauliString, pauli_literal_to_bools, pauli_product


def test_pauli_string_construction() -> None:
    ps1 = PauliString({0: "X", 1: "Y", 2: "Z"})
    ps2 = PauliString({0: "X", 2: "Z", 1: "Y"})
    empty = PauliString({})
    assert bool(ps1)
    assert not empty
    assert len(ps1) == 3
    assert ps1 == ps2
    assert len({ps1, ps2}) == 1
    assert ps1.overlaps(ps2)
    assert not ps1.overlaps(PauliString({3: "X"}))
    with pytest.raises(TQECDException, match=r"^Invalid Pauli operator.*"):
        PauliString({0: "W"})  # type: ignore


def test_pauli_string_interop_with_stim() -> None:
    stim_pauli_string = stim.PauliString.random(num_qubits=23)
    pauli_string = PauliString.from_stim_pauli_string(stim_pauli_string)
    assert (
        pauli_string.to_stim_pauli_string(length=23) * stim_pauli_string
    ).weight == 0

    pauli_string = PauliString.from_stim_pauli_string(stim.PauliString("_XYZ"))
    assert pauli_string.after(
        stim.Tableau.from_named_gate("CZ"), targets=[0, 1]
    ) == PauliString.from_stim_pauli_string(stim.PauliString("+ZXYZ"))

    assert PauliString.from_stim_pauli_string(stim.PauliString("IXYZ")) == PauliString(
        {0: "I", 1: "X", 2: "Y", 3: "Z"}
    )

    with pytest.raises(
        TQECDException, match=r"^The length specified 2 <= the maximum qubit index.*"
    ):
        pauli_string.to_stim_pauli_string(2)


def test_pauli_string_mul() -> None:
    a = PauliString({q: p for q, p in enumerate("IIIIXXXXYYYYZZZZ")})  # type:ignore
    b = PauliString({q: p for q, p in enumerate("IXYZ" * 4)})  # type:ignore
    c = PauliString({q: p for q, p in enumerate("IXYZXIZYYZIXZYXI")})  # type:ignore
    assert a * b == c


def test_pauli_string_commutation() -> None:
    a = PauliString({0: "X", 1: "Y"})
    b = PauliString({0: "Y", 1: "Z"})
    c = PauliString({0: "Z", 1: "Y"})
    assert a.commutes(b)
    assert b.commutes(a)
    assert a.anticommutes(c)
    assert c.anticommutes(a)
    assert b.commutes(c)
    assert c.commutes(b)


def test_pauli_string_collapse_by() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z"})
    Z0 = PauliString({0: "Z"})
    Z1 = PauliString({1: "Z"})
    X0 = PauliString({0: "X"})
    assert X0Z1.collapse_by([X0]) == Z1
    assert X0Z1.collapse_by([X0, Z1]) == PauliString({})
    with pytest.raises(TQECDException):
        X0Z1.collapse_by([Z0])


def test_pauli_string_weight() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z", 2: "I"})
    Z0 = PauliString({0: "Z"})
    I0to20 = PauliString({i: "I" for i in range(20)})
    assert X0Z1.non_trivial_pauli_count == 2
    assert Z0.non_trivial_pauli_count == 1
    assert I0to20.non_trivial_pauli_count == 0


def test_pauli_string_qubit() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z", 2: "I"})
    Z0 = PauliString({0: "Z"})
    I0to20 = PauliString({i: "I" for i in range(20)})
    assert Z0.qubit == 0
    with pytest.raises(TQECDException):
        X0Z1.qubit
    with pytest.raises(TQECDException):
        I0to20.qubit


def test_pauli_string_indexing() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z", 2: "I"})
    assert X0Z1[0] == "X"
    assert X0Z1[1] == "Z"
    assert X0Z1[2] == "I"
    assert X0Z1[3] == "I"


def test_pauli_literals_to_bool() -> None:
    assert pauli_literal_to_bools("I") == (False, False)
    assert pauli_literal_to_bools("X") == (True, False)
    assert pauli_literal_to_bools("Y") == (True, True)
    assert pauli_literal_to_bools("Z") == (False, True)


def test_pauli_product() -> None:
    X0Z1 = PauliString({0: "X", 1: "Z", 2: "I"})
    Z0 = PauliString({0: "Z"})
    I0to20 = PauliString({i: "I" for i in range(20)})

    assert pauli_product([I0to20 for _ in range(20)]) == I0to20
    assert pauli_product([X0Z1 for _ in range(20)]) == I0to20
    assert pauli_product([X0Z1 for _ in range(21)]) == X0Z1
    assert pauli_product([X0Z1, Z0]) == PauliString({0: "Y", 1: "Z"})



"""Automatic computation of detectors in quantum circuits representing fault-tolerant
quantum computations based on surface code and lattice surgery.

This package provides a set of functions and data-structures to help anyone
computing automatically the detectors (i.e., sets of deterministic measurements)
contained in a quantum circuit.
"""

from .construction import (
    annotate_detectors_automatically as annotate_detectors_automatically,
)



"""Introduces a convenience `PauliString` class.

This module implements an internal `PauliString` class with methods
that are used across the package. This class can easily be converted from
and to `stim.PauliString` and implement a subset of the `stim.PauliString`
API.
"""

from __future__ import annotations

import functools
import operator
from typing import Iterable, Literal

import stim

from tqecd.exceptions import TQECDException

PAULI_STRING_TYPE = Literal["I", "X", "Y", "Z"]
_IXYZ: list[PAULI_STRING_TYPE] = ["I", "X", "Y", "Z"]
_IXZY: list[PAULI_STRING_TYPE] = ["I", "X", "Z", "Y"]


class PauliString:
    """A mapping from qubits to Pauli operators that represent a Pauli string.

    Invariant:
        This class never stores identity Pauli terms. Any missing Pauli term is
        considered to be an identity.
        As such, it is illegal to initialise this class with an identity term.
    """

    def __init__(self, pauli_by_qubit: dict[int, PAULI_STRING_TYPE]) -> None:
        for qubit, pauli in pauli_by_qubit.items():
            if pauli not in _IXYZ:
                raise TQECDException(
                    f"Invalid Pauli operator {pauli} for qubit {qubit}, expected I, X, Y, or Z."
                )
        self._pauli_by_qubit: dict[int, PAULI_STRING_TYPE] = {
            q: pauli for q, pauli in sorted(pauli_by_qubit.items()) if pauli != "I"
        }
        self._hash = hash(tuple(self._pauli_by_qubit.items()))

    @property
    def non_trivial_pauli_count(self) -> int:
        return len(self._pauli_by_qubit)

    @property
    def qubits(self) -> Iterable[int]:
        return self._pauli_by_qubit.keys()

    @property
    def qubit(self) -> int:
        if len(self._pauli_by_qubit) != 1:
            raise TQECDException(
                "Cannot retrieve only one qubit from a Pauli string with "
                f"{len(self._pauli_by_qubit)} qubits."
            )
        return next(iter(self.qubits))

    @staticmethod
    def from_stim_pauli_string(
        stim_pauli_string: stim.PauliString,
    ) -> PauliString:
        """Convert a `stim.PauliString` to a `PauliString` instance, ignoring
        the sign."""
        return PauliString(
            {
                q: _IXYZ[stim_pauli_string[q]]
                for q in range(len(stim_pauli_string))
                if stim_pauli_string[q] != 0
            }
        )

    def to_stim_pauli_string(self, length: int | None) -> stim.PauliString:
        """Convert a `PauliString` to a `stim.PauliString` instance.

        Args:
            length: The length of the `stim.PauliString`. If `None`, the length is set to the
                maximum qubit index in the `PauliString` plus one.
        """
        max_qubit_index = max(self._pauli_by_qubit.keys())
        length = length or max_qubit_index + 1
        if length <= max_qubit_index:
            raise TQECDException(
                f"The length specified {length} <= the maximum qubit index {max_qubit_index} in the pauli string."
            )
        stim_pauli_string = stim.PauliString(length)
        for q, p in self._pauli_by_qubit.items():
            stim_pauli_string[q] = p
        return stim_pauli_string

    def __bool__(self) -> bool:
        return bool(self._pauli_by_qubit)

    def __mul__(self, other: PauliString) -> PauliString:
        result: dict[int, PAULI_STRING_TYPE] = {}
        for q in self._pauli_by_qubit.keys() | other._pauli_by_qubit.keys():
            a = self._pauli_by_qubit.get(q, "I")
            b = other._pauli_by_qubit.get(q, "I")
            ax = a in "XY"
            az = a in "YZ"
            bx = b in "XY"
            bz = b in "YZ"
            cx = ax ^ bx
            cz = az ^ bz
            c = _IXZY[cx + cz * 2]
            if c != "I":
                result[q] = c
        return PauliString(result)

    def __repr__(self) -> str:
        return f"PauliString(qubits={self._pauli_by_qubit!r})"

    def __str__(self) -> str:
        return "*".join(
            f"{self._pauli_by_qubit[q]}{q}" for q in sorted(self._pauli_by_qubit.keys())
        )

    def __len__(self) -> int:
        return len(self._pauli_by_qubit)

    def commutes(self, other: PauliString) -> bool:
        """Check if this Pauli string commutes with another Pauli string."""
        return not self.anticommutes(other)

    def anticommutes(self, other: PauliString) -> bool:
        """Check if this Pauli string anticommutes with another Pauli
        string."""
        t = 0
        for q in self._pauli_by_qubit.keys() & other._pauli_by_qubit.keys():
            t += self._pauli_by_qubit[q] != other._pauli_by_qubit[q]
        return t % 2 == 1

    def collapse_by(self, collapse_operators: Iterable[PauliString]) -> PauliString:
        """Collapse the provided Pauli string by the provided operators.

        Here, collapsing means that we are removing from the Pauli string represented
        by self all the commuting Pauli terms from all the provided operators.

        Collapsing is performed sequentially, in the order provided by
        `collapse_operators`. If, during this sequential collapsing, the current
        partially-collapsed result does not commute with the current collapsing
        operator, an exception is raised.

        Args:
            collapse_operators: a collection of operators that should all commute
                with self and will collapse with self.

        Raises:
            TQECDException: if one of the provided operators does not commute with self.

        Returns:
            a copy of self, collapsed by the provided operators.
        """
        ret = PauliString(self._pauli_by_qubit.copy())
        for op in collapse_operators:
            if not ret.commutes(op):
                raise TQECDException(
                    f"Cannot collapse {ret} by a non-commuting operator {op}."
                )
            for qubit in op.qubits:
                if qubit in ret._pauli_by_qubit:
                    del ret._pauli_by_qubit[qubit]
        return ret

    def after(self, tableau: stim.Tableau, targets: Iterable[int]) -> PauliString:
        stim_pauli_string = self.to_stim_pauli_string(
            length=max(list(targets) + list(self._pauli_by_qubit.keys())) + 1
        )
        stim_pauli_string_after = stim_pauli_string.after(tableau, targets=targets)
        return PauliString.from_stim_pauli_string(stim_pauli_string_after)

    def contains(self, other: PauliString) -> bool:
        return self._pauli_by_qubit.items() >= other._pauli_by_qubit.items()

    def overlaps(self, other: PauliString) -> bool:
        return bool(self._pauli_by_qubit.keys() & other._pauli_by_qubit.keys())

    def __eq__(self, other: object) -> bool:
        """Check if two PauliString are equal.

        Args:
            other: the instance to compare to.

        Returns:
            `True` if the two `PauliString` instances are equal, else False.
        """
        return (
            isinstance(other, PauliString)
            and self._pauli_by_qubit == other._pauli_by_qubit
        )

    def __hash__(self) -> int:
        return self._hash

    def __getitem__(self, index: int) -> PAULI_STRING_TYPE:
        return self._pauli_by_qubit.get(index, "I")


def pauli_literal_to_bools(
    literal: PAULI_STRING_TYPE,
) -> tuple[bool, bool]:
    if literal == "I":
        return (False, False)
    elif literal == "X":
        return (True, False)
    elif literal == "Y":
        return (True, True)
    elif literal == "Z":
        return (False, True)


def pauli_product(paulis: Iterable[PauliString]) -> PauliString:
    return functools.reduce(operator.mul, paulis, PauliString({}))



from typing import Iterable, Literal

import pytest
import stim

from tqecd.pauli import PauliString
from tqecd.utils import collapse_pauli_strings_at_moment, iter_stim_circuit_by_moments


def test_iter_by_moment_empty() -> None:
    circuit = stim.Circuit()
    assert len(list(iter_stim_circuit_by_moments(circuit))) == 0


def test_iter_by_moment_single_tick() -> None:
    circuit = stim.Circuit("TICK")
    assert len(list(iter_stim_circuit_by_moments(circuit))) == 1
    first_moment = next(iter_stim_circuit_by_moments(circuit))
    assert isinstance(first_moment, stim.Circuit)
    assert len(first_moment) == 1
    assert first_moment == circuit


def test_iter_by_moment_single_qec_round() -> None:
    circuit = stim.Circuit(
        """
        R 0 1 2 3 4
        TICK
        CX 0 1 2 3
        TICK
        CX 2 1 4 3
        TICK
        MR 1 3
        DETECTOR(1, 0) rec[-2]
        DETECTOR(3, 0) rec[-1]
        M 0 2 4
        DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
        DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
        OBSERVABLE_INCLUDE(0) rec[-1]"""
    )
    assert len(list(iter_stim_circuit_by_moments(circuit))) == 4
    moment_iterator = iter_stim_circuit_by_moments(circuit)
    moment = next(moment_iterator)
    assert isinstance(moment, stim.Circuit)
    assert len(moment) == 2
    assert moment == stim.Circuit("R 0 1 2 3 4\nTICK")
    moment = next(moment_iterator)
    assert isinstance(moment, stim.Circuit)
    assert len(moment) == 2
    assert moment == stim.Circuit("CX 0 1 2 3\nTICK")
    moment = next(moment_iterator)
    assert isinstance(moment, stim.Circuit)
    assert len(moment) == 2
    assert moment == stim.Circuit("CX 2 1 4 3\nTICK")
    moment = next(moment_iterator)
    assert isinstance(moment, stim.Circuit)
    assert len(moment) == 7
    assert moment == stim.Circuit(
        """MR 1 3
        DETECTOR(1, 0) rec[-2]
        DETECTOR(3, 0) rec[-1]
        M 0 2 4
        DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
        DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
        OBSERVABLE_INCLUDE(0) rec[-1]"""
    )


def test_iter_by_moment_repeat_block() -> None:
    circuit = stim.Circuit(
        """
        REPEAT 9 {
            TICK
            CX 0 1 2 3
            TICK
            CX 2 1 4 3
            TICK
            MR 1 3
            SHIFT_COORDS(0, 1)
            DETECTOR(1, 0) rec[-2] rec[-4]
            DETECTOR(3, 0) rec[-1] rec[-3]
        }"""
    )
    assert len(list(iter_stim_circuit_by_moments(circuit))) == 1
    moment = next(iter_stim_circuit_by_moments(circuit))
    assert isinstance(moment, stim.CircuitRepeatBlock)
    assert len(moment.body_copy()) == 9
    circuit = stim.Circuit(
        """
        R 0 1 2 3 4
        TICK
        CX 0 1 2 3
        TICK
        CX 2 1 4 3
        TICK
        MR 1 3
        REPEAT 9 {
            TICK
            CX 0 1 2 3
            TICK
            CX 2 1 4 3
            TICK
            MR 1 3
            SHIFT_COORDS(0, 1)
            DETECTOR(1, 0) rec[-2] rec[-4]
            DETECTOR(3, 0) rec[-1] rec[-3]
        }"""
    )
    # REPEAT block should NOT be with the last MR before it, even without TICK.
    assert len(list(iter_stim_circuit_by_moments(circuit))) == 5


@pytest.mark.parametrize("basis", ["", "X", "Y", "Z"])
@pytest.mark.parametrize("qubits", [range(5), [1, 2, 0], [0, 1, 2], [2, 1, 0]])
def test_collapse_pauli_strings_at_moment_all_basis(
    basis: Literal["", "X", "Y", "Z"], qubits: Iterable[int]
) -> None:
    qubits_str = " ".join(map(str, qubits))
    circuit = stim.Circuit(f"R{basis} {qubits_str}")
    collapsing_operations = collapse_pauli_strings_at_moment(circuit)
    pauli = basis or "Z"
    assert collapsing_operations == [PauliString({q: pauli}) for q in qubits]


def test_collapse_pauli_strings_at_moment_virtual_instruction() -> None:
    circuit = stim.Circuit("R 0 1 2\nX_ERROR(0.01) 0 1 2")
    collapsing_operations = collapse_pauli_strings_at_moment(circuit)
    assert set(collapsing_operations) == {PauliString({q: "Z"}) for q in [0, 1, 2]}



"""Semi-private module providing functions to construct SAT problems.

This module is semi-private in the sense that even though the two functions it
implements are not private, they are quite low-level and are not expected to be
used by external users of the library.

The module defines 2 functions that each take a SAT solver able to encode XOR
clauses, a target Pauli string, a list of potential Pauli strings and a set of
qubits to consider and encode in the provided SAT solver a cover problem, either
exact (i.e., find Pauli strings in the list of potential Pauli strings such that
their product is exactly the target Pauli string on the considered qubits) or
commuting (i.e., the product only needs to commute with the target Pauli string
on the considered qubits).
"""

from __future__ import annotations

from pysat.solvers import CryptoMinisat

from tqecd.pauli import PauliString, pauli_literal_to_bools


def encode_pauli_string_exact_cover_sat_problem_in_solver(
    solver: CryptoMinisat,
    expected_pauli_string: PauliString,
    available_pauli_strings: list[PauliString],
    qubits_to_consider: frozenset[int],
) -> None:
    """Build the SAT problem that should be solved to find an exact cover.

    This function encodes the SAT problem of interest into the provided ``solver``.
    As such, the provided ``solver`` is expected to be a newly created instance,
    and will be mutated by this function.

    The encoded problem is the following:

    Find the indices of Pauli strings in ``available_pauli_strings`` such that
    the product of all the corresponding Pauli strings is exactly equal to
    ``expected_pauli_string`` on all the qubits in ``qubits_to_consider``.

    The following post-condition should be checked after solving:

    .. code-block:: python

        expected_pauli_string = PauliString({})      # Fill in!
        available_pauli_strings = [PauliString({})]  # Fill in!
        qubits_to_consider = frozenset()             # Fill in!
        solver = None                                # Fill in!

        encode_pauli_string_exact_cover_sat_problem_in_solver(
            solver, expected_pauli_string, available_pauli_strings, qubits_to_consider
        )
        indices = None # Solve the problem encoded in solver and recover the indices.

        final_pauli_string = pauli_product([available_pauli_strings[i] for i in indices])
        for q in qubits_to_consider:
            assert final_pauli_string[q] == expected_pauli_string[q]


    The SAT problem contains one boolean variable per Pauli string in
    ``available_pauli_strings``, deciding if the associated Pauli string should
    be part of the cover.

    The formula is quite simple: for each qubit in ``qubits_to_consider``, the
    following two boolean formulas should be ``True``:

    1. The X part of the resulting Pauli string, obtained from XORing the X part
       of each Pauli string weighted by the boolean variable that decides if the
       Pauli string should be included or not, should be equal to the X part of
       the expected final Pauli string.
    2. Same reasoning, but for the Z part.

    Note that the "should be equal to the [P] part of the expected final Pauli
    string" is natively implemented in the SAT solver used, so we do not need to
    adapt the formula with some binary logic tricks. The solver really solves a
    conjunction of ``a XOR b XOR ... XOR z == [True/False]``.

    This leads to a formula with ``2*len(qubits_to_consider)`` XOR clauses that
    should all be verified for a cover to exist.
    Each of the ``2*len(qubits_to_consider)`` XOR clauses can contain up to
    ``len(available_pauli_strings)`` XORed items, but a simplification is made
    that should reduce that number: if the Pauli string acts trivially w.r.t the
    considered Pauli effect (X or Z), the boolean variable is removed from the
    formula as we know for sure that this particular Pauli string cannot impact
    the result on that specific qubit and Pauli effect.

    Args:
        solver: solver that will be modified in-place to encode the SAT problem.
        expected_pauli_string: target Pauli string that should be covered by
            strings from ``available_pauli_strings``.
        available_pauli_strings: Pauli strings that can be used to try to cover
            ``expected_pauli_string``.
        qubits_to_consider: qubits on which the cover should be exactly equal to
            ``expected_pauli_string``. Qubits not listed in this input will
            simply be ignored and no restriction on the value of the resulting
            cover on those qubits is added.
    """

    for qubit in qubits_to_consider:
        expected_X, expected_Z = pauli_literal_to_bools(expected_pauli_string[qubit])
        available_paulis_bools: list[tuple[bool, bool]] = [
            pauli_literal_to_bools(pauli[qubit]) for pauli in available_pauli_strings
        ]
        # The two following lists includes the 1-based indices of Pauli strings from
        # the provided `available_pauli_strings` input that finish with a X/Z stabilizers
        # on the current `qubit`.
        # For each index, we want to know if it should be included in the final cover or
        # not, so XORing the boolean variables representing whether or not this Pauli
        # string should be included results in the stabilizer propagated on that qubit.
        x_clause_literals = [
            pi + 1 for pi, (px, _) in enumerate(available_paulis_bools) if px
        ]
        z_clause_literals = [
            pi + 1 for pi, (_, pz) in enumerate(available_paulis_bools) if pz
        ]
        solver.add_xor_clause(x_clause_literals, value=expected_X)
        solver.add_xor_clause(z_clause_literals, value=expected_Z)


def encode_pauli_string_commuting_cover_sat_problem_in_solver(
    solver: CryptoMinisat,
    expected_pauli_string: PauliString,
    available_pauli_strings: list[PauliString],
    qubits_to_consider: frozenset[int],
) -> None:
    """Build the SAT problem that should be solved to find a commuting cover.

    This function encodes the SAT problem of interest into the provided ``solver``.
    As such, the provided ``solver`` is expected to be a newly created instance,
    and will be mutated by this function.

    The encoded problem is the following:

    Find the indices of Pauli strings in ``available_pauli_strings`` such that
    the product of all the corresponding Pauli strings commutes with
    ``expected_pauli_string`` on all the qubits in ``qubits_to_consider``.

    The SAT problem contains one boolean variable per Pauli string in
    ``available_pauli_strings``, deciding if the associated Pauli string should
    be part of the cover.

    The formula is quite simple: for each qubit in ``qubits_to_consider``, the
    following two boolean formulas should be ``True``:

    1. The X part of the resulting Pauli string, obtained from XORing the X part
       of each Pauli string weighted by the boolean variable that decides if the
       Pauli string should be included or not, should be equal to the X part of
       the expected final Pauli string or be equal to I.
    2. Same reasoning, but for the Z part.

    Note that the "should be equal to the [P] part of the expected final Pauli
    string" is natively implemented in the SAT solver used, so we do not need to
    adapt the formula with some binary logic tricks. The solver really solves a
    conjunction of ``a XOR b XOR ... XOR z == [True/False]``.

    Also note that the "or be equal to I" part is a little bit more tricky and
    requires some care. The logic retained for each of the possible Pauli terms
    is explained inline within the code.

    The number of XOR clauses in the formula depends on the number of each Pauli
    terms in the input ``expected_pauli_string`` that is considered (i.e. on
    qubits in ``qubits_to_consider``). Let [P] be the number of Pauli term equal
    to P in ``expected_pauli_string``, the number of XOR clauses in the
    resulting formula is ``(2 * [I] + [X] + [Y] + [Z]) * len(qubits_to_consider)``.
    That means that there are less clauses in the SAT problem defined in this
    function than in the one defined in
    :func:`encode_pauli_string_exact_cover_sat_problem_in_solver`, except if
    ``expected_pauli_string`` only contains identity gates, in which case the
    two functions return the same number of XOR clauses.

    Just like :func:`encode_pauli_string_exact_cover_sat_problem_in_solver`, the
    number of terms in each XOR clause can vary depending on the provided Pauli
    strings in ``available_pauli_strings``.

    Also note that the problem defined above includes a trivial solution: do not
    include any Pauli string. This lead to an identity Pauli string, that will
    necessarily commute with the provided target. A specific clause is added to
    the SAT problem to avoid that particular trivial solution.

    Args:
        solver: solver that will be modified in-place to encode the SAT problem.
        expected_pauli_string: target Pauli string that should be covered by
            strings from ``available_pauli_strings``.
        available_pauli_strings: Pauli strings that can be used to try to cover
            ``expected_pauli_string``.
        qubits_to_consider: qubits on which the cover should be exactly equal to
            ``expected_pauli_string``. Qubits not listed in this input will
            simply be ignored and no restriction on the value of the resulting
            cover on those qubits is added.
    """
    # Clause to exclude the trivial solution "include no Pauli string at all"
    solver.add_clause([lit + 1 for lit in range(len(available_pauli_strings))])

    for qubit in qubits_to_consider:
        expected_effect = expected_pauli_string[qubit]
        available_paulis_bools: list[tuple[bool, bool]] = [
            pauli_literal_to_bools(pauli[qubit]) for pauli in available_pauli_strings
        ]
        # The two following lists includes the 1-based indices of Pauli strings from
        # the provided `available_pauli_strings` input that finish with a X/Z stabilizers
        # on the current `qubit`.
        # For each index, we want to know if it should be included in the final cover or
        # not, so XORing the boolean variables representing whether or not this Pauli
        # string should be included results in the stabilizer propagated on that qubit.
        x_clause_literals = [
            pi + 1 for pi, (px, _) in enumerate(available_paulis_bools) if px
        ]
        z_clause_literals = [
            pi + 1 for pi, (_, pz) in enumerate(available_paulis_bools) if pz
        ]
        if expected_effect == "I":
            # Both the X and Z effect on that qubit should be OFF (i.e., identity).
            solver.add_xor_clause(x_clause_literals, value=False)
            solver.add_xor_clause(z_clause_literals, value=False)
        elif expected_effect == "X":
            # The X effect should be ON, the Z effect should be OFF.
            # Because the identity (0, 0) is also valid, X can be either ON or OFF,
            # so we do not need to restrict X effect.
            solver.add_xor_clause(z_clause_literals, value=False)
        elif expected_effect == "Y":
            # This is an expected Y effect. This means that X and Z should be both
            # ON (i.e., the Y effect) or both OFF (i.e., the identity effect).
            # Rephrasing using XOR, this is equivalent to `X_effect XOR Z_effect == 0`.
            # Note that measurements appearing twice here can be removed, as:
            # - the XOR operation is commutative (so we can re-organise the measurements
            #   to be sorted by index).
            # - b XOR b = 0.
            # - b XOR 0 = b.
            # Also, a given measurement cannot appear more than twice, so we are fine
            # just removing duplicated entries.
            solver.add_xor_clause(
                list(set(x_clause_literals) ^ set(z_clause_literals)), value=False
            )
        elif expected_effect == "Z":
            # The X effect should be OFF, the Z effect should be ON.
            # Because the identity (0, 0) is also valid, Z can be either ON or OFF,
            # so we do not need to restrict Z effect.
            solver.add_xor_clause(x_clause_literals, value=False)



import pytest
import stim

from tqecd.exceptions import TQECDException
from tqecd.fragment import Fragment, FragmentLoop, split_stim_circuit_into_fragments
from tqecd.pauli import PauliString


def test_creation() -> None:
    Fragment(stim.Circuit("""M 1 2 3 4 5"""))
    Fragment(stim.Circuit("R 1 2 3\nTICK\nM 1 2 3"))
    Fragment(
        stim.Circuit("""
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
M 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
M 0 2 4
DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
OBSERVABLE_INCLUDE(0) rec[-1]""")
    )
    Fragment(stim.Circuit("QUBIT_COORDS(0, 0) 0\nTICK\nM 1 2 3"))
    with pytest.raises(TQECDException):
        Fragment(stim.Circuit("R 1 2 3"))
    with pytest.raises(TQECDException):
        Fragment(
            stim.Circuit("""
REPEAT 10 {
    R 0 1 2 3 4
    TICK
    CX 0 1 2 3
    TICK
    CX 2 1 4 3
    TICK
    M 1 3
    DETECTOR(1, 0) rec[-2]
    DETECTOR(3, 0) rec[-1]
    M 0 2 4
    DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
    DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
    OBSERVABLE_INCLUDE(0) rec[-1]
}""")
        )
    with pytest.raises(TQECDException):
        Fragment(stim.Circuit("M 1 2 3\nR 1 2 3"))


def test_measurement_qubits() -> None:
    fragment = Fragment(stim.Circuit("M 1 2 3"))
    assert set(fragment.measurements_qubits) == {1, 2, 3}
    assert fragment.num_measurements == 3


def test_fragment_measurements_and_resets() -> None:
    fragment = Fragment(stim.Circuit("M 1 2 3"))
    assert len(fragment.resets) == 0
    assert set(fragment.measurements) == {PauliString({i: "Z"}) for i in [1, 2, 3]}

    fragment = Fragment(stim.Circuit("R 1 2 3\nTICK\nM 1 2 3"))
    assert set(fragment.measurements) == {PauliString({i: "Z"}) for i in [1, 2, 3]}
    assert set(fragment.measurements) == {PauliString({i: "Z"}) for i in [1, 2, 3]}

    fragment = Fragment(
        stim.Circuit("""
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
M 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
M 0 2 4
DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
OBSERVABLE_INCLUDE(0) rec[-1]""")
    )
    assert set(fragment.resets) == {PauliString({i: "Z"}) for i in range(5)}
    assert set(fragment.measurements) == {PauliString({i: "Z"}) for i in range(5)}


def test_get_tableau() -> None:
    measurement_only = Fragment(stim.Circuit("""M 0 1 2 """))
    reset_measurement_only = Fragment(stim.Circuit("""R 0 1 2\nTICK\nM 0 1 2"""))

    assert measurement_only.get_tableau() == stim.Tableau(3)
    assert reset_measurement_only.get_tableau() == stim.Tableau(3)

    h_gates = Fragment(stim.Circuit("""H 0 1 2\nTICK\nM 0 1 2 """))
    assert h_gates.get_tableau() == stim.Circuit("H 0 1 2").to_tableau()


def test_fragment_loop_creation() -> None:
    fragment = Fragment(stim.Circuit("R 1 2 3\nTICK\nM 1 2 3"))
    FragmentLoop([fragment], 10)
    FragmentLoop([fragment for _ in range(10)], 10)
    # Even though a single repetition is not really something that makes sense (why using
    # a REPEAT in that case?), it is functionally valid, so we should accept it.
    FragmentLoop([fragment], 1)
    with pytest.raises(TQECDException):
        FragmentLoop([], 10)
    with pytest.raises(TQECDException):
        FragmentLoop([fragment], 0)
    with pytest.raises(TQECDException):
        FragmentLoop([fragment], -10)


def test_fragment_loop_with_repetition() -> None:
    fragment = Fragment(stim.Circuit("R 1 2 3\nTICK\nM 1 2 3"))
    fragment_loop = FragmentLoop([fragment], 10)
    assert fragment_loop.with_repetitions(1) == FragmentLoop([fragment], 1)
    assert fragment_loop.with_repetitions(10) == fragment_loop
    with pytest.raises(TQECDException):
        fragment_loop.with_repetitions(0)
    with pytest.raises(TQECDException):
        fragment_loop.with_repetitions(-10)


def test_split_stim_circuit_into_fragments_simple() -> None:
    fragments = split_stim_circuit_into_fragments(stim.Circuit("M 0 1 2"))
    assert len(fragments) == 1
    assert fragments[0] == Fragment(stim.Circuit("M 0 1 2"))

    fragments = split_stim_circuit_into_fragments(
        stim.Circuit("R 1 2 3\nTICK\nM 1 2 3")
    )
    assert len(fragments) == 1
    assert fragments[0] == Fragment(stim.Circuit("R 1 2 3\nTICK\nM 1 2 3"))

    one_round_qec = stim.Circuit("""
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
M 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
M 0 2 4
DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
OBSERVABLE_INCLUDE(0) rec[-1]""")
    fragments = split_stim_circuit_into_fragments(one_round_qec)

    assert len(fragments) == 1
    assert fragments[0] == Fragment(one_round_qec)


def test_split_stim_circuit_into_fragments_qec_repetition_code() -> None:
    circuit = stim.Circuit("""
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
M 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
TICK
REPEAT 9 {
    R 1 3
    TICK
    CX 0 1 2 3
    TICK
    CX 2 1 4 3
    TICK
    M 1 3
    SHIFT_COORDS(0, 1)
    DETECTOR(1, 0) rec[-2] rec[-4]
    DETECTOR(3, 0) rec[-1] rec[-3]
    TICK
}
M 0 2 4
DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
OBSERVABLE_INCLUDE(0) rec[-1]""")
    first_round = stim.Circuit("""
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
M 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
TICK""")
    repeat_body = stim.Circuit("""
R 1 3
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
M 1 3
SHIFT_COORDS(0, 1)
DETECTOR(1, 0) rec[-2] rec[-4]
DETECTOR(3, 0) rec[-1] rec[-3]
TICK""")
    # No reset in end_round.
    # The last TICK is an implementation detail, maybe try to remove it?
    end_round = stim.Circuit("""
M 0 2 4
DETECTOR(1, 1) rec[-2] rec[-3] rec[-5]
DETECTOR(3, 1) rec[-1] rec[-2] rec[-4]
OBSERVABLE_INCLUDE(0) rec[-1]""")
    fragments = split_stim_circuit_into_fragments(circuit)
    assert len(fragments) == 3
    assert fragments[0] == Fragment(first_round)
    assert fragments[1] == FragmentLoop([Fragment(repeat_body)], 9)
    assert fragments[2] == Fragment(end_round)


def test_split_stim_circuit_into_fragments_repeat_block() -> None:
    circuit = stim.Circuit("""
REPEAT 9 {
    R 1 3
    TICK
    M 1 3
}""")
    fragments = split_stim_circuit_into_fragments(circuit)
    assert len(fragments) == 1
    assert fragments[0] == FragmentLoop(
        [Fragment(stim.Circuit("R 1 3\nTICK\nM 1 3"))], 9
    )

    erroneous_circuit = stim.Circuit("REPEAT 9 {\nR 1 3\nTICK\nH 1 3\n}")
    with pytest.raises(
        TQECDException, match=r"^Error when splitting .* REPEAT block.*"
    ):
        split_stim_circuit_into_fragments(erroneous_circuit)


def test_split_stim_circuit_into_fragments_error_before_repeat() -> None:
    circuit = stim.Circuit("""
R 1 3
REPEAT 9 {
    R 1 3
    TICK
    M 1 3
}""")
    with pytest.raises(
        TQECDException,
        match=r"Trying to start a REPEAT block without a cleanly finished Fragment.*",
    ):
        split_stim_circuit_into_fragments(circuit)



from __future__ import annotations

import itertools
from typing import Iterator

import pytest
import stim

from tqecd.match_utils.cover import (
    _all_pauli_string_combination_results,  # pyright: ignore[reportPrivateUsage]
    find_commuting_cover_on_target_qubits_sat,
    find_exact_cover_sat,
)
from tqecd.pauli import PauliString, pauli_product


def _pss(pauli_string: str) -> PauliString:
    return PauliString.from_stim_pauli_string(stim.PauliString(pauli_string))


def _all_pauli_string_combination_results_alternative_implementation(
    pauli_string_list: list[PauliString],
) -> Iterator[PauliString]:
    yield from (
        pauli_product([p for p, c in zip(pauli_string_list, choices) if c])
        for choices in itertools.product([True, False], repeat=len(pauli_string_list))
    )


def test_all_pauli_string_combination_results_randomized() -> None:
    paulis = [
        PauliString.from_stim_pauli_string(stim.PauliString.random(10))
        for _ in range(5)
    ]

    from_function = {
        pauli for _, pauli in _all_pauli_string_combination_results(paulis)
    }
    from_alternative_implementation = set(
        _all_pauli_string_combination_results_alternative_implementation(paulis)
    )
    assert from_function == from_alternative_implementation


def test_all_pauli_string_combination_results() -> None:
    paulis = [PauliString({i: "Z"}) for i in range(3)]
    all_combinations = list(
        pauli for _, pauli in _all_pauli_string_combination_results(paulis)
    )
    assert len(all_combinations) == 2 ** len(paulis)
    assert set(all_combinations) == {
        PauliString({}),
        PauliString({0: "Z"}),
        PauliString({1: "Z"}),
        PauliString({2: "Z"}),
        PauliString({0: "Z", 1: "Z"}),
        PauliString({0: "Z", 2: "Z"}),
        PauliString({1: "Z", 2: "Z"}),
        PauliString({0: "Z", 1: "Z", 2: "Z"}),
    }


@pytest.mark.parametrize(
    "target,sources,expected_result",
    [
        (
            _pss("__ZZ__ZZ__"),
            [PauliString({i: "Z"}) for i in range(10)],
            [2, 3, 6, 7],
        ),
        (
            _pss("__ZZ__ZZ__"),
            [PauliString({i: "X"}) for i in range(10)],
            None,
        ),
        (
            _pss("YYYZZZ"),
            [PauliString({i: "Z"}) for i in range(6)]
            + [PauliString({i: "X"}) for i in range(6)],
            [0, 1, 2, 3, 4, 5, 6, 7, 8],
        ),
        (
            _pss("_XYZ_XYZ"),
            [PauliString({i: "Z"}) for i in range(8)]
            + [PauliString({i: "X"}) for i in range(8)],
            [2, 3, 6, 7, 9, 10, 13, 14],
        ),
        (
            _pss("_XYZ_XYZ"),
            [PauliString({i: "Z"}) for i in range(7)]  # Missing the last Z
            + [PauliString({i: "X"}) for i in range(8)],
            None,
        ),
        (
            _pss("____"),
            [PauliString({i: "Z"}) for i in range(4)]
            + [PauliString({i: "X"}) for i in range(4)],
            [],
        ),
        (
            _pss("YYYY"),
            [PauliString({i: "Z", (i + 1) % 4: "X"}) for i in range(4)],
            [0, 1, 2, 3],
        ),
        (
            _pss("X"),
            [],
            None,
        ),
        (
            _pss("_"),
            [],
            [],
        ),
    ],
)
def test_exact_match(
    target: PauliString, sources: list[PauliString], expected_result: list[int] | None
) -> None:
    obtained_result = find_exact_cover_sat(target, sources)
    # We expect the results to either both be None, or both be a list.
    assert (obtained_result is None) == (expected_result is None)
    # If they are both a list, compare them and check the post-condition documented.
    if obtained_result is not None and expected_result is not None:
        assert set(obtained_result) == set(expected_result)
        assert pauli_product([sources[i] for i in obtained_result]) == target


@pytest.mark.parametrize(
    "target,sources,expected_result",
    [
        (
            _pss("X"),
            [],
            None,
        ),
        (
            _pss("_"),
            [],
            None,
        ),
        (
            _pss("ZZ_XX_ZZ"),
            [_pss("ZZZZZZZZ"), _pss("___YY___"), _pss("XXXXXXXX")],
            [0, 1],
        ),
        (
            _pss("ZZ_YY_ZZ"),
            [_pss("ZZZZZZZZ"), _pss("XX____XX"), _pss("XXXXXXXX")],
            [0, 1, 2],
        ),
    ],
)
def test_commuting_match(
    target: PauliString, sources: list[PauliString], expected_result: list[int] | None
) -> None:
    obtained_result = find_commuting_cover_on_target_qubits_sat(target, sources)
    # We expect the results to either both be None, or both be a list.
    assert (obtained_result is None) == (expected_result is None)
    # If they are both a list, compare them and check the post-condition documented.
    if obtained_result is not None and expected_result is not None:
        assert set(obtained_result) == set(expected_result)
        assert pauli_product([sources[i] for i in obtained_result]).commutes(target)



from __future__ import annotations

from typing import Iterable, Mapping

import numpy

from tqecd.exceptions import TQECDException
from tqecd.measurement import RelativeMeasurementLocation
from tqecd.pauli import PauliString


class BoundaryStabilizer:
    def __init__(
        self,
        stabilizer: PauliString,
        collapsing_operations: Iterable[PauliString],
        measurements: list[RelativeMeasurementLocation],
        reset_qubits: frozenset[int],
        forward: bool,
    ):
        """Represents a stabilizer that has been propagated and is now at the
        boundary of a :class:`~tqecd.fragment.Fragment`.

        Raises:
            TQECDException: if ``source_qubits`` is empty.

        Args:
            stabilizer: The propagated stabilizer **before** any collapsing
                operation has been applied.
            collapsing_operations: The collapsing operations the stabilizer will
                have to go through to exit the :class:`~tqecd.fragment.Fragment`.
            measurements: measurement offsets relative to the **end** of the
                fragment (even if the created :class:`BoundaryStabilizer`
                instance represents a stabilizer on the beginning boundary) of
                measurements that are involved in this stabilizer.
            reset_qubits: index of the qubit on which reset operations touching
                the flow are applied. Depending on the value of ``forward``,
                these indices can either be sources of the stabilizer (if
                ``forward`` is ``True``) or sinks that may or may not commute
                with the resulting stabilizer.
            forward:
                ``True`` if the stabilizer propagated forward (i.e., ends on
                measurements), else ``False``.
        """
        self._stabilizer = stabilizer
        self._measurements = measurements
        self._collapsing_operations = frozenset(collapsing_operations)
        self._has_anticommuting_collapsing_operations = any(
            co.anticommutes(stabilizer) for co in collapsing_operations
        )
        self._reset_qubits: frozenset[int] = reset_qubits
        self._is_forward = forward

    @property
    def has_anticommuting_operations(self) -> bool:
        """Check if the instance represents a stabilizer that anti-commutes
        with at least one of its collapsing operations.

        Returns:
            ``True`` if at least one collapsing operation anti-commutes with the
            stabilizer, else ``False``.
        """
        return self._has_anticommuting_collapsing_operations

    @property
    def after_collapse(self) -> PauliString:
        """Compute the stabilizer obtained after applying the collapsing
        operations.

        Raises:
            TQECDException: If any of the collapsing operation anti-commutes
                with the stored stabilizer.

        Returns:
            The collapsed :class:`~tqecd.pauli.PauliString` that goes out of the
            :class:`~tqecd.fragment.Fragment`.
        """
        if self.has_anticommuting_operations:
            raise TQECDException(
                "Cannot collapse a BoundaryStabilizer if it has "
                "anticommuting operations."
            )
        return self._stabilizer.collapse_by(self._collapsing_operations)

    @property
    def before_collapse(self) -> PauliString:
        """Return the stabilizer obtained before applying the collapsing
        operations.

        Returns:
            The :class:`~tqecd.pauli.PauliString` that goes out of the
            :class:`~tqecd.fragment.Fragment`, before applying any collapsing
            operation.
        """
        return self._stabilizer

    @property
    def collapsing_operations(self) -> Iterable[PauliString]:
        """Iterator on all the collapsing operations defining the boundary this
        stabilizer is applied to."""
        return self._collapsing_operations

    @property
    def measurements(self) -> list[RelativeMeasurementLocation]:
        return self._measurements

    @property
    def resets_qubits(self) -> frozenset[int]:
        return self._reset_qubits

    def merge(self, other: BoundaryStabilizer) -> BoundaryStabilizer:
        """Merge two boundary stabilizers together.

        The two merged stabilizers should be defined on the same boundaries. In
        particular, they should have the same set of collapsing operations.

        Args:
            other: the other :class:`BoundaryStabilizer` to merge with ``self``.
                Should have exactly the same set of collapsing operations.

        Raises:
            TQECDException: if ``self`` and ``other`` are not defined on exactly
                the same collapsing operations.

        Returns:
            the merged boundary stabilizer, defined on the same set of collapsing
            operations (i.e., the same boundary), but with the two pre-collapsing
            stabilizers multiplied together.
        """
        self_collapsing_operations = set(self.collapsing_operations)
        other_collapsing_operations = set(other.collapsing_operations)
        if self_collapsing_operations != other_collapsing_operations:
            raise TQECDException(
                "Breaking pre-condition: trying to merge two BoundaryStabilizer "
                "instances that are not defined on the same boundary.\n"
                f"Collapsing operations for left-hand side: {self_collapsing_operations}.\n"
                f"Collapsing operations for right-hand side: {other_collapsing_operations}.\n"
            )
        if self._is_forward != other._is_forward:
            raise TQECDException(
                "Cannot merge a forward boundary stabilizer with a backward one."
            )
        is_forward_merge = self._is_forward
        stabilizer = self._stabilizer * other._stabilizer
        non_trivial_stabilizer_qubits = frozenset(stabilizer.qubits)
        # When merging 2 boundary stabilizers, particular care should be taken to merge
        # the measurements and resets.
        # The merged stabilizers might cancel out on some of the involved collapsing
        # operations, leading to collapsing operations that are in `self` and `other`
        # but that should not be in the merged result.
        # This does not happen everytime a collapsing operation is found in both
        # `self` and `other` as:
        # - merging anti-commuting stabilizer (which is the primary purpose of this
        #   method) might lead to a commuting stabilizer, still involving the collapsing
        #   operation they were previously anti-commuting with.
        # - sources (i.e., resets in forward boundary stabilizers and measurements in
        #   backward boundary operations) do not cancel each other out.
        # In the end, the end collapsing operations (resets for backward propagation,
        # measurements for forward propagation) need to be re-computed from the
        # merged stabilizer.
        reset_qubits: frozenset[int]
        measurements: list[RelativeMeasurementLocation]
        if is_forward_merge:
            reset_qubits = self.resets_qubits | other.resets_qubits
            candidate_measurements = set(self.measurements) | set(other.measurements)
            measurements = [
                m
                for m in candidate_measurements
                if m.qubit_index in non_trivial_stabilizer_qubits
            ]

        else:
            measurements = list(
                frozenset(self.measurements) | frozenset(other.measurements)
            )
            candidate_resets = self.resets_qubits | other.resets_qubits
            reset_qubits = frozenset(
                r for r in candidate_resets if r in non_trivial_stabilizer_qubits
            )
        return BoundaryStabilizer(
            stabilizer,
            self_collapsing_operations,
            measurements,
            reset_qubits,
            is_forward_merge,
        )

    def __repr__(self) -> str:
        ret = f"BoundaryStabilizers(stabilizer={self._stabilizer}, "
        ret += "collapsing_operations=["
        ret += ", ".join(str(p) for p in self.collapsing_operations)
        ret += f"], measurements={self._measurements}"
        ret += f", resets={set(self._reset_qubits)}, is_forward={self._is_forward})"
        return ret

    def coordinates(
        self, qubit_coordinates: Mapping[int, tuple[float, ...]]
    ) -> tuple[float, ...]:
        """Compute and return the coordinates of the boundary stabilizer.

        The coordinates of a given boundary stabilizer is defined as the average
        of the coordinates of each collapsing operations it represents.

        Args:
            qubit_coordinates: mapping from qubit indices to coordinates

        Raises:
            TQECDException: If a qubit in ``self.source_qubits`` is not contained
                in the ``qubit_coordinates`` mapping.

        Returns:
            the boundary stabilizer coordinates.
        """
        try:
            measurement_coordinates = [
                qubit_coordinates[source] for source in self.source_qubits
            ]
        except KeyError as exc:
            raise TQECDException(
                f"Qubit index {exc.args[0]} required for detector assignment, "
                "but it does not have a valid QUBIT_COORDS statement."
            )
        # Avoid returning numpy.float64 type returned by numpy.mean by
        # explicitly calling float() on it.
        return tuple(float(c) for c in numpy.mean(measurement_coordinates, axis=0))

    def with_measurement_offset(self, offset: int) -> BoundaryStabilizer:
        return BoundaryStabilizer(
            self._stabilizer,
            self.collapsing_operations,
            [m.offset_by(offset) for m in self.measurements],
            self._reset_qubits,
            self._is_forward,
        )

    def is_trivial(self) -> bool:
        return (
            not self.has_anticommuting_operations
            and self.after_collapse.non_trivial_pauli_count == 0
            and len(self._stabilizer) == 1
        )

    @property
    def source_qubits(self) -> frozenset[int]:
        return (
            self._reset_qubits
            if self._is_forward
            else frozenset(m.qubit_index for m in self.measurements)
        )


def manhattan_distance(
    lhs: BoundaryStabilizer,
    rhs: BoundaryStabilizer,
    qubit_coordinates: dict[int, tuple[float, ...]],
) -> float:
    lhs_coords = lhs.coordinates(qubit_coordinates)
    rhs_coords = rhs.coordinates(qubit_coordinates)
    return sum(abs(left - right) for left, right in zip(lhs_coords, rhs_coords))



from __future__ import annotations

import typing as ty

import stim

from tqecd.utils import (
    has_measurement,
    has_reset,
    is_combined_measurement_reset,
    iter_stim_circuit_by_moments,
)


def _has_any_combined_gates(circuit: stim.Circuit) -> bool:
    for instruction in circuit:
        if isinstance(instruction, stim.CircuitRepeatBlock):
            if _has_any_combined_gates(instruction.body_copy()):
                return True
        elif is_combined_measurement_reset(instruction):
            return True
    return False


def does_not_contain_combined_gates(circuit: stim.Circuit) -> bool:
    return not _has_any_combined_gates(circuit)


def does_not_contain_both_reset_and_measurement(circuit: stim.Circuit) -> bool:
    for moment in iter_stim_circuit_by_moments(circuit):
        if isinstance(moment, stim.CircuitRepeatBlock):
            if not does_not_contain_both_reset_and_measurement(moment.body_copy()):
                return False
        else:
            if has_measurement(moment) and has_reset(moment):
                return False
    return True


def is_valid_input_circuit(circuit: stim.Circuit) -> str | None:
    predicates: list[tuple[ty.Callable[[stim.Circuit], bool], str]] = [
        (
            does_not_contain_combined_gates,
            "The provided quantum circuit should not contain any combined instruction (e.g., MR).",
        ),
        (
            does_not_contain_both_reset_and_measurement,
            "The provided quantum circuit contains at least one moment that has both reset "
            "and measurement operations.",
        ),
    ]

    for pred, reason in predicates:
        if not pred(circuit):
            return reason
    return None
